#include <vortex/acpi/acpiheaders.h>
#include <vortex/acpi/platform/acpi_vortex.h>
#include <vortex/acpi/platform/vortex_acpi_pstates.h>
#include "../core/arch/x86_64/cpu.h"
#include <vortex/cpumux.h>


static mutex_t pstate_register_mutex;

/******************************************************************************
 *
 * FUNCTION:     acpi_init_acpi_pstate_info
 *
 * PARAMETERS:  cpuid       The cpuid of the core whose acpi based pstate 
 *                          information is to be initialized
 *
 * RETURN:      valid pointer		Initialization successfull.
 *              NULL                Initialization failed.
 *
 * DESCRIPTION: This function is used to initialize all P-state information 
 *              presented by ACPI for a given core.
 *
 *****************************************************************************/
acpi_pstate_info_t *acpi_init_acpi_pstate_info(UINT8 cpuid)
{
    acpi_pstate_info_t  *acpi_pstate_info;
    vxerr_t             vxerr;

    assert(cpuid >= 0 && cpuid < CPU_DRAFTNUM);

    acpi_pstate_info = (acpi_pstate_info_t *)pm_alloc(sizeof(acpi_pstate_info_t), VX_FALSE);
    if (acpi_pstate_info == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate acpi_pstate_info structure");
        goto error;
    }

    acpi_pstate_info->api_acpi_pstate_array = acpi_get_performance_supported_states(cpuid);
    if (acpi_pstate_info->api_acpi_pstate_array == NULL) {
        syslog(LOG_CRIT, "Unable to retrieve supported power states for cpu %d from ACPI", cpuid);
        pm_free((vaddr_t)acpi_pstate_info, sizeof(acpi_pstate_info_t));
        goto error;
    }

    vxerr = acpi_retrieve_performance_control_registers(cpuid, &acpi_pstate_info->api_use_ffh, &acpi_pstate_info->api_control_register, &acpi_pstate_info->api_status_register);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to retrieve performance (p-state) control registers for cpu %d from ACPI", cpuid);
        pm_free((vaddr_t)acpi_pstate_info, sizeof(acpi_pstate_info_t));
        goto error;
    }

    return acpi_pstate_info;

error:
    return NULL;
}

/******************************************************************************
 *
 * FUNCTION:     acpi_get_max_available_performance_state
 *
 * PARAMETERS:  cpuid       The cpuid of the core whose current maximum 
 *                          pstate is to be retrieved.
 *
 * RETURN:      max_pstate > 0		Maximum pstate sucessfully retrieved.
 *              -1                  cpuid out of range or internal error.
 *
 * DESCRIPTION: This function is used to obtain currently max performance 
 *              state for a given cpu core.
 *
 *****************************************************************************/
int32 acpi_get_max_available_performance_state(UINT8 cpuid)
{
	const utf8_t	*OSCPaths[8] = {"\\_PR_.CPU0._PPC", "\\_PR_.CPU1._PPC", "\\_PR_.CPU2._PPC", "\\_PR_.CPU3._PPC", "\\_PR_.CPU4._PPC", "\\_PR_.CPU5._PPC", "\\_PR_.CPU6._PPC", "\\_PR_.CPU7._PPC"};
	int32		    max_available_pstate;

	if (cpuid < 0 || cpuid > 7) {
		syslog(LOG_DEBUG, "ERROR: In acpi_get_max_available_performance_state - cpuid out of range");
		return -1;
	}

	max_available_pstate = execute_PPC(OSCPaths[cpuid]);
	if (max_available_pstate < 0)
		syslog(LOG_DEBUG, "Return value from execute_PPC is -1");

	return max_available_pstate;
}

/******************************************************************************
 *
 * FUNCTION:     acpi_get_performance_supported_states
 *
 * PARAMETERS:  cpuid       The cpuid of the core whose current maximum 
 *                          pstate is to be retrieved.
 *
 * RETURN:      pointer to struct of ACPI pstates supported by the core.		
 *              NULL        Non-valid cpuid, internal errorn
 *
 * DESCRIPTION: This function is used to obtain a struct of all available 
 *              (ACPI enumerable) pstates for the given core. Note that this 
 *              struct should be freed using a pm_free() call and employing 
 *              the trick used when we have structs ending with an array of 
 *              size 0.
 *
 *****************************************************************************/
pstate_array_t *acpi_get_performance_supported_states(UINT8 cpuid)
{
	const utf8_t			*_PSSPaths[8] = {"\\_PR_.CPU0._PSS", "\\_PR_.CPU1._PSS", "\\_PR_.CPU2._PSS", "\\_PR_.CPU3._PSS", "\\_PR_.CPU4._PSS", "\\_PR_.CPU5._PSS", "\\_PR_.CPU6._PSS", "\\_PR_.CPU7._PSS"};
	ACPI_OBJECT				*object, *current;
	UINT32					i;
	pstate_array_t		    *p_state_array;


	if (cpuid < 0 || cpuid > 7) {
		syslog(LOG_DEBUG, "ERROR: In acpi_get_performance_supported_states - cpuid out of range");
		return NULL;
	}

    object = execute_acpi_object_with_return_type_package(_PSSPaths[cpuid]);
	if (object == NULL) {
		syslog(LOG_DEBUG, "package is NULL, execution failed for object %s ", _PSSPaths[cpuid]);
		return NULL;
	}
	
	p_state_array = (pstate_array_t *)pm_alloc(sizeof(pstate_array_t) + sizeof(pstate_t) * object->Package.Count, VX_FALSE);
	p_state_array->num_elems = object->Package.Count;
	
	//Copy the pstate information
	for (i = 0; i < object->Package.Count; i++) {
		current = &object->Package.Elements[i];	
		p_state_array->elems[i].core_frequency = current->Package.Elements[0].Integer.Value;
		p_state_array->elems[i].power = current->Package.Elements[1].Integer.Value;
		p_state_array->elems[i].latency = current->Package.Elements[2].Integer.Value;
		p_state_array->elems[i].bus_master_latency = current->Package.Elements[3].Integer.Value;
		p_state_array->elems[i].control = current->Package.Elements[4].Integer.Value;
		p_state_array->elems[i].status = current->Package.Elements[5].Integer.Value;
	}

	AcpiOsFree(object);

	return p_state_array;
}

vxerr_t acpi_retrieve_performance_control_registers(UINT8 cpuid, uint32 *use_ffh, uint64 *control_register, uint64 *status_register)
{
	const utf8_t			*_PCTPaths[8] = {"\\_PR_.CPU0._PCT", "\\_PR_.CPU1._PCT", "\\_PR_.CPU2._PCT", "\\_PR_.CPU3._PCT", "\\_PR_.CPU4._PCT", "\\_PR_.CPU5._PCT", "\\_PR_.CPU6._PCT", "\\_PR_.CPU7._PCT"};
	ACPI_OBJECT				*object;
    vxerr_t                 vxerr;

	//Execute the _PCT method for the cpu - obtaining the object/
    object = execute_acpi_object_with_return_type_package(_PCTPaths[cpuid]);
	if (object == NULL) {
		syslog(LOG_DEBUG, "package is NULL, execution failed for object %s ", _PCTPaths[cpuid]);
        vxerr = VXERR_NEXIST;
        goto error;
	}

	//Validate the object
	if ( object->Package.Count != 2 || object->Type != ACPI_TYPE_PACKAGE) {
		syslog(LOG_DEBUG, "ERROR: Package returned from executing %s has incorrect format", _PCTPaths[cpuid]);
        vxerr = VXERR_INTERNAL;
        goto error;
	}

	//Get information about control register
	if ( object->Package.Elements[0].Type == ACPI_TYPE_BUFFER && object->Package.Elements[0].Buffer.Length >= sizeof(ffh_gas_t)) {
		
        if ( object->Package.Elements[0].Buffer.Pointer[3] == 0x1) {
			*use_ffh = 0;
			*control_register = (uint64)object->Package.Elements[0].Buffer.Pointer[7];

		} else if (object->Package.Elements[0].Buffer.Pointer[3] == 0x7F) {
			*use_ffh = 1;
			*control_register = (uint64)0x199;
		}
	
    } else {
		syslog(LOG_DEBUG, "ERROR: Control register object returned from _PCT is invalid");
        vxerr = VXERR_INTERNAL;
        goto error;
	}

	//Get information about status register
	if ( object->Package.Elements[1].Type == ACPI_TYPE_BUFFER && object->Package.Elements[1].Buffer.Length >= sizeof(ffh_gas_t)) {
		
        if ( object->Package.Elements[1].Buffer.Pointer[3] == 0x1)
			*status_register = (uint64)object->Package.Elements[1].Buffer.Pointer[7];
		
		else if (object->Package.Elements[1].Buffer.Pointer[3] == 0x7F)
			*status_register = (uint64)0x199;
	
    } else {
		syslog(LOG_DEBUG, "ERROR: Control register object returned from _PCT is invalid");
		vxerr = VXERR_INTERNAL;
        goto error;
	}
	AcpiOsFree(object);
    vxerr = VXERR_OK;

error:
    return vxerr;
}

int enable_enhanced_speedstep(void)
{
	uint64	msr_value;
	//Should set bit 16 of MSR_IA32_MISC_ENABLE (0x1A0)
	msr_value = asm_read_msr(0x1A0);

	syslog(LOG_DEBUG, "IA32_MISC_ENABLE %x", msr_value);
	//Set the 16'th bit
	msr_value |= 0x0000000000010000;
	syslog(LOG_DEBUG, "About to write IA32_MISC_ENABLE %x", msr_value);
	//Write back to register
	asm_write_msr(0x1A0, msr_value);

	msr_value = asm_read_msr(0x1A0);

	syslog(LOG_DEBUG, "After writing: IA32_MISC_ENABLE %x", msr_value);

	return 0;
}


// This function should be used from the core local thread.
vxerr_t acpi_change_performance_state(UINT8 cpuid, pstate_array_t *pstate_array, uint16 desired_performance_state, uint32 use_ffh, uint64 control_register, uint64 status_register)
{
	uint32	control_value, status_value, status, delay;
    vxerr_t vxerr;
    lock_t  lock;
	
	if (pstate_array == NULL) {
		syslog(LOG_CRIT, "ERROR: In acpi_change_performance_state - pstate_array pointer is NULL");
        vxerr = VXERR_ILARG;
        goto error;
	}

	/*Check that the desired power state is not lower than anything supported by the CPU*/
	if (desired_performance_state >= pstate_array->num_elems)
		desired_performance_state = pstate_array->num_elems -1;

	/*Obtain the control value for transitioning into the desired state*/
	control_value = pstate_array->elems[desired_performance_state].control;

	/*Obtain the status value for verifying the transition into the desired state*/
	status_value = pstate_array->elems[desired_performance_state].status;

	/*Obtain the delay*/
	delay = pstate_array->elems[desired_performance_state].latency;

	/*Perfrom writing to correct register*/
	if (use_ffh)
	{
		syslog(LOG_DEBUG, "ERROR: As our platform does not support Intel Enhanced Speedstep - neither do we ;(");
        vxerr = VXERR_ILARG;
        goto error;
	
    } else  {

        VxO_LOCKNULL(&lock);
        VxO_LOCKM(&lock, &pstate_register_mutex);
        AcpiOsWritePort (control_register, control_value, 8);
        VxO_UNLOCK(&lock);
		arch_delay_usec(delay);

		/*Validate transition*/
        AcpiOsReadPort (status_register, &status, 8);
        if (status == status_value) 
		    vxerr = VXERR_OK;
	    else
            vxerr = VXERR_INTERNAL;
	}

error:
    return vxerr;
}