#include <vortex/acpi/acpiheaders.h>

static int bitwidth_supported(UINT32 width);
static int pci_bitwidth_supported(UINT32 width);
/*
 * Platform and hardware-independent I/O interfaces
 */

/******************************************************************************
 *
 * FUNCTION:     AcpiOsReadPort
 *
 * PARAMETERS:  Address				- hardware I/O port address to read from.
 *              Value               - A pointer to a location where the data is 
 *									  to be returned.
 *				Width				- The port width in bits, either 8, 16, or 
 *									  32.
 *
 * RETURN:      Status		Exception code that indicates success or reason for 
 *							failure.
 *
 * DESCRIPTION: This function reads data from the specified input port. The 
 *				data is zero extended to fill the 32-bit return value even if 
 *				the bit width of the port is less than 32.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsReadPort (ACPI_IO_ADDRESS Address, UINT32 *Value, UINT32 Width)
{
	if ( Address == (ACPI_IO_ADDRESS)NULL || Value == NULL || !bitwidth_supported(Width))
		return AE_BAD_PARAMETER;

	switch ((uint32)Width) 
	{
		case 8:
			*Value = inb((int)Address);
			break;
		case 16:
			*Value = inw((int)Address);
			break;
		case 32:
			*Value = inl((int)Address);
			break;
	}

	return AE_OK;
}

/******************************************************************************
 *
 * FUNCTION:     AcpiOsWritePort
 *
 * PARAMETERS:  Address				- hardware I/O port address to be writeen to.
 *              Value               - The value to be written.
 *				Width				- The port width in bits, either 8, 16, or 
 *									  32.
 *
 * RETURN:      Status		Exception code that indicates success or reason for 
 *							failure.
 *
 * DESCRIPTION: This function writes data to the spcified input port. If the bit 
 *				width of the port is less than 32, only the lower significant 
 *				bits of the Value parameter are written.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsWritePort (ACPI_IO_ADDRESS Address, UINT32 Value, UINT32 Width)
{
	if ( Address == (ACPI_IO_ADDRESS)NULL || !bitwidth_supported(Width))
		return AE_BAD_PARAMETER;

	switch ((uint32)Width) 
	{
		case 8:
			outb((int)Value, (int)Address);
			break;
		case 16:
			outw((int)Value, (int)Address);
			break;
		case 32:
			outl((int)Value, (int)Address);
			break;
		}
  
	return AE_OK;
}


/*
 * Platform and hardware-independent PCI configuration space access
 * Note: Can't use "Register" as a parameter, changed to "Reg" --
 * certain compilers complain.
 */

/******************************************************************************
 *
 * FUNCTION:     AcpiOsReadPciConfiguration
 *
 * PARAMETERS:  PciId				- The full PCI configuration space address, 
 *									  consisting of a segment numberm bus number, 
 *									  device number, and a function number.
 *              Register            - The PCI register address to be read from.
 *				Value				- A pointer to a location where the data is 
 *									  to be returned.
 *				Width				- The register width in bits, either 8, 16, 
 *									  32, or 64.
 *
 * RETURN:      Status		Exception code that indicates success or reason for 
 *							failure.
 *
 * DESCRIPTION: This function reads data from the specified PCI configuration 
 *				port. The data is zero extended to fill the 64-bit return 
 *				value even if the bit wisth of the location is less than 64.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsReadPciConfiguration (ACPI_PCI_ID *PciId, UINT32 Reg, UINT64 *Value, UINT32 Width)
{
	if ( PciId == (ACPI_PCI_ID *)NULL || !pci_bitwidth_supported(Width) || Reg == (UINT32)NULL)
		return AE_BAD_PARAMETER;

	*Value = _pci_read_config( (uint16)PciId->Function, (uint16)Reg, (uint8)((uint32)Width / 8));
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:     AcpiOsWritePciConfiguration
 *
 * PARAMETERS:  PciId				- The full PCI configuration space address, 
 *									  consisting of a segment numberm bus number, 
 *									  device number, and a function number.
 *              Register            - The PCI register address to be written to.
 *				Value				- Data to be written.
 *				Width				- The register width in bits, either 8, 16, 
 *									  32, or 64.
 *
 * RETURN:      Status		Exception code that indicates success or reason for 
 *							failure.
 *
 * DESCRIPTION: This function writes data to the specified PCI configuration 
 *				port. If the bit width of the register is less than 64, only 
 *				the lower significant bits of the Value are written.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsWritePciConfiguration (ACPI_PCI_ID *PciId, UINT32 Reg, UINT64 Value, UINT32 Width)
{   
	vxerr_t status;

	if ( PciId == NULL || !pci_bitwidth_supported(Width) || Reg == (UINT32)NULL)
		return AE_BAD_PARAMETER;

	status = _pci_write_config( (uint64)Value, (uint16)PciId->Function, (uint16)Reg, (uint8)((uint32)Width / 8));

	if (status == VXERR_OK)
		return AE_OK;
	else
		return AE_ERROR;
}


/*Simply check if the bitwidth is supported.*/
static int bitwidth_supported(UINT32 width)
{
	int isOK = 0;
	switch((uint32)width)
	{
		case 8:
			isOK = 1;
			break;
		case 16:
			isOK = 1;
			break;
		case 32:
			isOK = 1;
			break;
		default:
			break;
	}
	
	return isOK;
}

static int pci_bitwidth_supported(UINT32 width)
{
	if (bitwidth_supported(width))
		return 1;
	else
	{
		if ((uint32)width == 64)
			return 1;
	}

	return 0;
}

