#include <vortex/acpi/acpiheaders.h>

#define NUM_SUPPORTED_INTERRUPTS 10	 

typedef void (*VORTEX_INTERRUPT_HANDLER_CALLBACK)(uint32_t interrupt_number); //Will be innecessary

vxerr_t osd_interrupt_handler(excdetail_t *exeption_detail);

static int is_valid(uint32_t interrupt_number);

/*Struct for storing acppi interrupts*/
struct acpi_interrupt {						
	UINT32				ai_InterruptNumber;
	ACPI_OSD_HANDLER	ai_ServiceRoutine;
	void				*ai_Context;
};

/*Create a Vortex kernel object type for stored acpica interrupt*/
type_t acpimemAcpiInterrupt_type = {
    .t_name = "acpimemAcpiInterrupt",
    .t_size = sizeof(acpi_interrupt_t),
    .t_alloc = KOBJ_ZEROFILL,
};

/*Table of stored acpi interrupts*/
dict_t *stored_acpi_interrupts;
uint32_t interrupt_storage_ready = 0;								


/******************************************************************************
 *
 * FUNCTION:     AcpiOsInstallInterruptHandler
 *
 * PARAMETERS:  InterruprNumber	 - Interrupt number that handler will service.
 *				Handler		     - Address of the handler.
 *				Context			 - A context value that is passed to the handler 
 *								   when the interrupt is dispatched. 
 *
 * RETURN:      AE_OK				The handler was successfully installed.
 *				AE_BAD_PARAMETER	Interrupt number invalid or Handler is NULL.
 *				AE_ALREADY_EXISTS	A handler for this interrupt is already 
 *									installed.
 *
 * DESCRIPTION: This function installs an interrupt handler for a hardware 
 *				interrupt level/number. The ACPI driver must install an 
 *				interrupt handler to service the SCI (System Control Interrupt) 
 *				which it owns. The interrupt level for SCI interrupt is 
 *				obtained from the ACPI tables.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsInstallInterruptHandler (UINT32 InterruptNumber, ACPI_OSD_HANDLER ServiceRoutine, void *Context)
{
    acpi_interrupt_t	*interrupt;
	vxerr_t				vxerr;
	objref_t			interruptref;

    /*Sanity check of input*/
    if (!is_valid(InterruptNumber) || ServiceRoutine == NULL)
        return AE_BAD_PARAMETER;
    
    /*Initialize the storage if this is first time this function is run*/
    if (!interrupt_storage_ready)
    {
        init_interrupt_storage();
    }

	/*Check if key allready present, if so return AE_ALREADY_EXISTS*/
	vxerr = VxD_GETITEM(&interruptref, stored_acpi_interrupts, (uid_t)InterruptNumber, NULL);
    if (vxerr == VXERR_OK)
        goto key_present_error;

	/*Create the interrupt object to be stored*/
	interrupt = (acpi_interrupt_t *)VxO_NEW(acpimemAcpiInterrupt);
    interrupt->ai_InterruptNumber = (uint32)InterruptNumber;
    interrupt->ai_Context = Context;
    interrupt->ai_ServiceRoutine = ServiceRoutine;
	
	/*Register the interrupt in the kernel*/
	vxerr = _arch_exception_handler_set(InterruptNumber,  osd_interrupt_handler);
	syslog(LOG_DEBUG, "Managed to install interrupt handler for Interrupt vector %d ", InterruptNumber);

	/*Store the interrupt object in the dict*/
	VxO_REFOBJ(&interruptref, interrupt);
	vxerr = VxD_SETITEM(&interruptref, stored_acpi_interrupts, (uid_t)InterruptNumber, NULL);
	if (vxerr != VXERR_OK)
		goto insertion_error;
		
	return AE_OK;

key_present_error:

	/*While testing - halt on failure*/
	syslog(LOG_CRIT, "Interrupt Handler for key %d already present", InterruptNumber);
	return AE_ALREADY_EXISTS;

insertion_error:
	/*While testing - halt on failure*/
	syslog(LOG_CRIT, "Interrupt Handler for key %d could not be installed", InterruptNumber);

	VxO_XDECREF(interrupt);
	return AE_ERROR;
}

/******************************************************************************
 *
 * FUNCTION:     AcpiOsRemoveInterruptHandler
 *
 * PARAMETERS:  InterruprNumber	 - Interrupt number that handler is currently 
 *								   servicing.
 *				Handler		     - Address of the handler that was previously 
 *								   installed.
 *
 * RETURN:      AE_OK				The handler was successfully removed.
 *				AE_BAD_PARAMETER	Interrupt number invalid or Handler is NULL.
 *									Or, the handler address is not the same 
 *									as the one that is installed.
 *				AE_NOT_EXISTS		There is no handler for this interrupt 
 *									level.
 *
 * DESCRIPTION: Remove a previously installed interrupt handler.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsRemoveInterruptHandler (UINT32 InterruptNumber, ACPI_OSD_HANDLER ServiceRoutine)
{
    acpi_interrupt_t	*stored_acpi_interrupt;
	vxerr_t				vxerr;
	objref_t			interruptref;

    /*Sanity check input*/
    if (!is_valid(InterruptNumber) || ServiceRoutine == NULL)
        return AE_BAD_PARAMETER;

	/*Start by locating the interrupt*/
	vxerr = VxD_GETITEM(&interruptref, stored_acpi_interrupts, (uid_t)InterruptNumber, NULL);
    if (vxerr != VXERR_OK)
        goto key_not_present_error;

	stored_acpi_interrupt = (acpi_interrupt_t *) VxO_REFGETOBJ(&interruptref);

	//Check if the supplied Handler is the same as the one stored//
    if (stored_acpi_interrupt->ai_ServiceRoutine != ServiceRoutine)
    {
        return AE_BAD_PARAMETER;
    }
    else
    {
        //Remove the stored interrupt handler//
		vxerr = VxD_DELITEM(&interruptref, stored_acpi_interrupts, (uid_t)InterruptNumber, NULL);
        if (vxerr != VXERR_OK)
            goto deletion_error;

		VxO_DECREF(stored_acpi_interrupt);
        return AE_OK;
    }  

key_not_present_error:
	/*While testing - halt on failure*/
	syslog(LOG_CRIT, "Interrupt Handler for key %d could not be deleted - Key not present", InterruptNumber);

	return AE_NOT_EXIST;

deletion_error:
	/*While testing - halt on failure*/
	syslog(LOG_CRIT, "Interrupt Handler for key %d could not be deleted - Internal Error", InterruptNumber);
	
	return AE_ERROR;
}


/*Helper and home grown functions*/


/******************************************************************************
 *
 * FUNCTION:     osd_interrupt_handler
 *
 * PARAMETERS:  interrupt_number	- Number corresponding to a interrupt. The 
 *									  interrupt number the that will be used to 
 *									  look up a handler that will be dispatched.
 *
 * RETURN:      None
 *
 * DESCRIPTION: A OSL specific wrapper function. This function will be passed 
 *				as argument (function pointer) to the host OS kernel, and it 
 *				will be invoked whenever a interrupt subscribed to is caught 
 *				by the host OS interrupt handler.
 *
 *				Example:	The ACPICA subsystem uses the 
 *							AcpiOsInstallInterruptHandler function to install
 *							a handler that they supply with corresponding 
 *							context, binding these two to a interrupt number.
 *
 *							The OSL layer will make a call to the host OS 
 *							kernel, registering this function together with the
 *							given interrupt number. The OSL will now subscribe 
 *							to this interrupt number, and this function will be
 *							run whenever the host OS interrupt handler catches 
 *							an interrupt with a number to which this function 
 *							is associated.
 *
 *							When this function is run (host OS catches an 
 *							interrupt we subscribe to), the interrupt number 
 *							is passed as parameter. This function now performs 
 *							a lookup with this interrupt number, and invokes 
 *							the ACPICA handler installed for this interrupt 
 *							together with the context associated with it.
 *
 *****************************************************************************/
vxerr_t osd_interrupt_handler(excdetail_t *exeption_detail)
{
    uint32				interrupt_number;
    acpi_interrupt_t	*stored_acpi_interrupt;
    ACPI_OSD_HANDLER	handler;
	vxerr_t				vxerr;
	objref_t			interruptref;

	interrupt_number = exeption_detail->ed_vector;
	vxerr = VxD_GETITEM(&interruptref, stored_acpi_interrupts, (uid_t)interrupt_number, NULL);
    if (vxerr != VXERR_OK)
        goto key_not_present_error;

	stored_acpi_interrupt = (acpi_interrupt_t *) VxO_REFGETOBJ(&interruptref);
	handler = stored_acpi_interrupt->ai_ServiceRoutine;
	handler(stored_acpi_interrupt->ai_Context);
    
	return VXERR_OK;

key_not_present_error:
	syslog(LOG_CRIT, "No handler for interrupt %d found \n", interrupt_number);
	return VXERR_NEXIST;
}



/******************************************************************************
 *
 * FUNCTION:     init_interrupt_storage
 *
 * PARAMETERS:  None
 *
 * RETURN:      None
 *
 * DESCRIPTION: Initializes the interrupt storage, enabling the OSL to install 
 *				handlers and contexts for interrupts. This function should be 
 *				called when initializing the OSL, e.g. from the 
 *				AcpiOsInitialize function.
 *
 *****************************************************************************/
ACPI_STATUS init_interrupt_storage()
{
	stored_acpi_interrupts = VxD_NEW(SYNC_GLOBAL, DICT_TYPE_IKEY);
	if (stored_acpi_interrupts == NULL)
	{
        syslog(LOG_CRIT, "Unable to create iterrupt storage dict");
		return;
	}

    interrupt_storage_ready = 1;
	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:     is_valid
 *
 * PARAMETERS:  interrupt_number	- Interrupt number to be tested for validity.
 *
 * RETURN:      1	The interrupt number was found to be valid.
 *				2	The interrupt number was found to be invalid.
 *
 * DESCRIPTION: Checks interrupt number for validity. Used internally in the 
 *				OSL interrupt handler functions.
 *
 *****************************************************************************/
static int is_valid(UINT32 interrupt_number)
{
    if (interrupt_number >= 0 && interrupt_number < 100)
        return 1;
    else
        return 0;
}
