#include <vortex/acpi/acpiheaders.h>

//BEGIN Types
/*Object types corresponding to sizes that are predefined*/
type_t acpimemACPI_SIZE_type = {
    .t_name = "acpimemACPI_SIZE",
    .t_size = sizeof(ACPI_SIZE),
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimemACPI_MEMORY_LIST_type = {
    .t_name = "acpimemACPI_MEMORY_LIST",
    .t_size = sizeof(ACPI_MEMORY_LIST),
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimemACPI_DEBUG_BUFFER_SIZE_type = {
    .t_name = "acpimemACPI_DEBUG_BUFFER",
    .t_size = ACPI_DEBUG_BUFFER_SIZE,
    .t_alloc = KOBJ_ZEROFILL,
};


/*Object types not corresponding to exact sizes*/
type_t acpimem32_type = {
    .t_name = "acpimem32",
    .t_size = 32,
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimem64_type = {
    .t_name = "acpimem64",
    .t_size = 64,
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimem128_type = {
    .t_name = "acpimem128",
    .t_size = 128,
    .t_alloc = KOBJ_ZEROFILL,
};


type_t acpimem256_type = {
    .t_name = "acpimem256",
    .t_size = 256,
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimem512_type = {
    .t_name = "acpimem512",
    .t_size = 512,
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimem768_type = {
    .t_name = "acpimem768",
    .t_size = 768,
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpimem1024_type = {
    .t_name = "acpimem1024",
    .t_size = 1024,
    .t_alloc = KOBJ_ZEROFILL,
};
type_t acpimem1280_type = {
    .t_name = "acpimem1280",
    .t_size = 1280,
    .t_alloc = KOBJ_ZEROFILL,
};
type_t acpimem2048_type = {
    .t_name = "acpimem2048",
    .t_size = 2048,
    .t_alloc = KOBJ_ZEROFILL,
};

/*
 * Memory allocation and mapping functions for use by ACPICA subsystem
 */


/******************************************************************************
 *
 * FUNCTION:    AcpiOsAllocate
 *
 * PARAMETERS:  Size                Amount to allocate, in bytes
 *
 * RETURN:      Pointer to the new allocation.  Null on error.
 *
 * DESCRIPTION: Allocate memory.  Algorithm is dependent on the OS.
 *
 *****************************************************************************/
void *AcpiOsAllocate (ACPI_SIZE Size)
{

	/* In Vortex, we do not allocate memory in the way we are used to (malloc). Instead, 
	 * we have to create data types of the sizes we wish to allocate later. When we try 
	 * to allocate memory corresponding to the size of such a type, we will actually get 
	 * back an entire object (with headers etc.) but the returned pointer will point 
	 * directly to the memory buffer of the desired size and is thus ready for use. 
	 *
	 * We can use this as a basis and subtract ourselves back so that we can read out 
	 * the headers info.
	 */

	/*Have a set of predefined exact sizes (see the types part at the top of the file)*/

	void *memory = NULL;
	
	switch(Size) {
	case sizeof(ACPI_SIZE): // This is a UINT64
		memory = VxO_NEW(acpimemACPI_SIZE);
		break;
	case sizeof(ACPI_MEMORY_LIST): //This is a struct
		memory = VxO_NEW(acpimemACPI_MEMORY_LIST);
		break;
	case ACPI_DEBUG_BUFFER_SIZE: //This is an actual number
		memory = VxO_NEW(acpimemACPI_DEBUG_BUFFER_SIZE);
		break;
	}
	
	/*Request did not match any exact objects, find most fitting type*/
	if (memory == NULL)
	{
		
		if (Size <= 32) {
			memory = VxO_NEW(acpimem32);
		} else if (Size <= 64) {
			memory = VxO_NEW(acpimem64);
		} else if (Size <= 128) {
			memory = VxO_NEW(acpimem128);
		} else if (Size <= 256) {
			memory = VxO_NEW(acpimem256);
		} else if (Size <= 512) {
			memory = VxO_NEW(acpimem512);
		} else if (Size <= 768) {
			memory = VxO_NEW(acpimem768);
		} else if (Size <= 1024) {
			memory = VxO_NEW(acpimem1024);
		} else if (Size <= 1280) {
			memory = VxO_NEW(acpimem1280);
		} else if (Size <= 2048) {
			memory = VxO_NEW(acpimem2048);
		} else
			syslog(LOG_CRIT, "Requested memory size from AcpiOsAllocate not supported (%d bytes)\n", (int)Size);
	}

	if (memory == NULL)
		syslog(LOG_CRIT, "Vx_NEW have given us NULL !!!!");

	return memory;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsFree
 *
 * PARAMETERS:  Memory			Pointer to previously allocated memory.
 *
 * RETURN:      None.
 *
 * DESCRIPTION: Free memory allocated via AcpiOsAllocate.
 *
 *****************************************************************************/
void AcpiOsFree (void *Memory)
{
	/* In vortex, we do not really free or alloc memory as we are used to. Instead, 
	 * we simply decrease the number of references to this piece of memory. When the 
	 * number of references is zero, this means that it is unreferences and can be 
	 * reused. This is handled by the memory subsystem.
	 */
	if (Memory != NULL)
		VxO_DECREF(Memory);
	else
		syslog(LOG_DEBUG, "AcpiOsFree: Attempted to free NULL pointer.\n");
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsMapMemory
 *
 * PARAMETERS:  Where               A full physical address of the memory to be 
 *									mapped into the caller's address space.
 *              Length              The amount of memory to be mapped starting
 *									at the given physical address.
 *
 * RETURN:      Pointer to mapped memory.  Null on error.
 *
 * DESCRIPTION: Map physical memory into caller's address space. A logical 
 *				pointer is returned.
 *
 *****************************************************************************/
void *AcpiOsMapMemory (ACPI_PHYSICAL_ADDRESS Where, ACPI_SIZE Length)
{
	return (void *)PTOV((uint64)Where);
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsUnmapMemory
 *
 * PARAMETERS:  LogicalAddress      Logical address that was returned from a 
 *									previous call to AcpiOsMapMemory.
 *              Length              The amount of memory that was mapped. This 
 *									value must be identical to the value used 
 *									the call to AcpiOsMapMemory.
 *
 * RETURN:      None.
 *
 * DESCRIPTION: Delete a previously created mapping.  Where and Length must
 *              correspond to a previous mapping exactly.
 *
 *****************************************************************************/
void AcpiOsUnmapMemory (void *LogicalAddress, ACPI_SIZE Size)
{
	/*We have no paging, nothing to do.*/
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsGetPhysicalAddress
 *
 * PARAMETERS:  LogicalAddress      Logical address to be translated
 *              PhysicalAddress     The physical memory address of the logical 
 *									address.
 *
 * RETURN:      AE_OK				The logical address translation was 
 *									succesfull.
 *				AE_ERROR			An error occured in translation system call.
 *				AE_BAD_PARAMETER	One or both of the parameters are NULL, no
 *									translation was attempted.
 *
 * DESCRIPTION: This function translates a logical address to its physical 
 *				address location.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsGetPhysicalAddress (void *LogicalAddress, ACPI_PHYSICAL_ADDRESS *PhysicalAddress)
{
	if (PhysicalAddress == NULL || LogicalAddress == NULL)
		return AE_BAD_PARAMETER;

	*PhysicalAddress = (ACPI_PHYSICAL_ADDRESS)VTOP((uint64)LogicalAddress);

	return AE_OK;
}

/*
 * Platform and hardware-independent physical memory interfaces
 */

/******************************************************************************
 *
 * FUNCTION:    AcpiOsReadMemory
 *
 * PARAMETERS:  Address             Physical Memory Address to be read.
 *              Value               A pointer to a location where the data is 
 *									to be returned.
 *              Width               The memory widt in bits, either 8, 16, or 32.
 *
 * RETURN:      AE_OK				Memory successfully read.
 *				AE_BAD_PARAMETER	Inputs are NULL or unnsupported.
 *
 * DESCRIPTION: This function is used to read a data from the specified memory 
 *				location. The data is zero-extended to fill the 32-bit return 
 *				value even if the bit width of the location is less than 32. 
 *				In other words, a full 32 bits are written to the return Value 
 *				pointer regardless of the number of bits that were read from the 
 *				memory ad Address. The caller must ensure that no data will be 
 *				overwritten by this call.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsReadMemory (ACPI_PHYSICAL_ADDRESS Address, UINT64 *Value, UINT32 Width)
{
	if (Address == (ACPI_PHYSICAL_ADDRESS)NULL || Value == (UINT64 *)NULL)
		return AE_BAD_PARAMETER;

	switch ((uint32)Width) 
	{
		case 8:
			*Value = *(volatile uint8 *)Address;
			break;
		case 16:
			*Value = *(volatile uint16 *)Address;
			break;
		case 32:
			*Value = *(volatile uint32 *)Address;
			break;
		case 64:
			*Value = *(volatile uint64 *)Address;
			break;
	}
	
	return AE_OK;
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsWriteMemory
 *
 * PARAMETERS:  Address             Physical Memory Address to write
 *              Value               Value to write
 *              Width               Number of bits
 *
 * RETURN:      
 *
 * DESCRIPTION: Write data to a physical memory address
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsWriteMemory (ACPI_PHYSICAL_ADDRESS Address, UINT64 Value, UINT32 Width)
{
	if (Address == (ACPI_PHYSICAL_ADDRESS)NULL)
		return AE_BAD_PARAMETER;

	switch ((uint32)Width) 
	{
		case 8:
			*(volatile uint8_t *)Address = Value;
			break;
		case 16:
			*(volatile uint16_t *)Address = Value;
			break;
		case 32:
			*(volatile uint32_t *)Address = Value;
			break;
		case 64:
			*(volatile uint64_t *)Address = Value;
			break;
	}

	return AE_OK;
}

/*
 * Miscellaneous
 */

/******************************************************************************
 *
 * FUNCTION:    AcpiOsReadable
 *
 * PARAMETERS:  Pointer             - Area to be verified
 *              Length              - Size of area
 *
 * RETURN:      TRUE if readable for entire length
 *
 * DESCRIPTION: Verify that a pointer to a memory region is valid and the entire
 *				region is readable. Used to validate input parameters to the 
 *				ACPICA subsystem.
 *
 *****************************************************************************/
BOOLEAN AcpiOsReadable (void *Pointer, ACPI_SIZE Length)
{
	/*In Vortex everything is readable*/
	return (TRUE);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsWritable
 *
 * PARAMETERS:  Pointer             - Area to be verified
 *              Length              - Size of area
 *
 * RETURN:      TRUE if the entire region is both readable and writable 
 *				without faults.
 *				FALSE if one or more bytes within the region are unreadable 
 *				or unwritable.
 *
 * DESCRIPTION: Verify that a pointer to a memory region is valid and the entire
 *				region is both writable and readable. Used to validate input 
 *				paramters to the ACPICA subsystem.
 *
 *****************************************************************************/
BOOLEAN AcpiOsWritable (void *Pointer, ACPI_SIZE Length)
{
	/*If we are not in the kernel code, everything is writable*/
	return TRUE;
}