#include <vortex/acpi/acpiheaders.h>

objref_t    tcbref;
objref_t	queueref;
rqueue_t    *scheduler_queue;

/*Helper functions*/
static void acpi_scheduler_thread(void);
static int execution_type_is_ok(ACPI_EXECUTE_TYPE type);



/******************************************************************************
 *
 * FUNCTION:     AcpiOsThreadId
 *
 * PARAMETERS:  None
 *
 * RETURN:      ThreadId	A unique non-zero value that represents the ID of 
 *							the currently executing thread. For single threaded 
 *							implementations, a constant integer > zero is 
 *							acceptable. The value 0xFFFFFFFFFFFFFFFF (-1) is 
 *							reserves and must not be returned by this interface.
 *
 * DESCRIPTION: This function returns the ID of the currently executing thread. 
 *				The value must be non-zero and must be unique to the executing 
 *				thread. The ACPI_THREAD_ID is an unsigned, 64-bit value. It is 
 *				up to the host OSL to cast the native thread ID to an 
 *				ACPI_THREAD_ID.
 *
 *				A 64-bit ACPI_THREAD_ID is used since it is the only data type 
 *				that can be used to handle all of the various native thread ID
 *				types (32-bit integer, 64-bit integer, 32-bit pointer, 64-bit 
 *				pointer).
 *
 *****************************************************************************/
ACPI_THREAD_ID AcpiOsGetThreadId (void)
{
	return (ACPI_THREAD_ID)VxO_UID(CURRENT_THREAD);
}


/******************************************************************************
 *
 * FUNCTION:     AcpiOsExecute
 *
 * PARAMETERS:  Type		- Type of callback function: OSL_GLOBAL_LOCK_HANDLER,
 *							  OSL_NOTIFY_HANDLER, OSL_GPE_HANDLER, 
 *							  OSL_DEBUGGER_THREAD, OSL_EC_POLL_HANDLER, 
 *							  OS_EC_BURST_HANDLER
 *				Function	- Address of the procedure to execute.
 *				Context		- A context value to be passed to the called 
 *							  procedure.
 *
 * RETURN:      AE_OK				The procedure was sucessfully queued for 
 *									execution by the host operating system. 
 *									This does not indicate that the procedure 
 *									has actually executed, however.
 *				AE_BAD_PARAMETER	Priority is invalid or the Function pointer 
 *									is NULL.
 *
 * DESCRIPTION: The function queues a procedure for later scheduling and 
 *				execution.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsExecute (ACPI_EXECUTE_TYPE Type, ACPI_OSD_EXEC_CALLBACK Function, void *Context)
{
	closure_t	*closure;
	vxerr_t		vxerr;

	if (!execution_type_is_ok(Type) || Function == NULL)
		return AE_BAD_PARAMETER;

	//Initialize
	closure = (closure_t *)NULL;
	
	//Try to create a new closure containing task and arguments
	closure = closure_new(Function, "p",Context);
	if (closure == NULL)
        goto error;

	//Enqueue the created closure
	vxerr = _rq_enqueue(&queueref, REQ_USERDEFINED, closure, NULL, (cookietype_t)NULL);
	if (vxerr != VXERR_OK)
        goto error;

	return AE_OK;

error:
    VxO_XDECREF(closure);
	return AE_NO_MEMORY;
}


/******************************************************************************
 *
 * FUNCTION:     AcpiOsWaitEventsComplete
 *
 * PARAMETERS:  None
 *
 * RETURN:      None
 *
 * DESCRIPTION: This interface is defined to block until asynchronous events 
 *				such as notifies and GPEs have completed. Within ACPICA, it is 
 *				only called before a notify or GPE handler is 
 *				removed/uninstalled. It also may be useful for the host OS 
 *				within related drivers such as the Embedded Controller driver. 
 *				See the ACPICA reference for additional information. 
 *				ACPICA BZ 868.
 *
 *****************************************************************************/
void AcpiOsWaitEventsComplete (void)
{
	time_t sleeptime;
	
	sleeptime = 10000; //10 milliseconds
	while(scheduler_queue->rq_len > 0 )
	{
		thread_suspend_timeout(NULL, sleeptime);
	}
}


/******************************************************************************
 *
 * FUNCTION:     AcpiOsSleep
 *
 * PARAMETERS:  Milliseconds		The amount of time to sleep, in millisecs.
 *
 * RETURN:      None
 *
 * DESCRIPTION: The function sleeps for the specified time. Execution of the 
 *				running thread is suspended for this time. The sleep 
 *				granularity is one millisecond.
 *
 *****************************************************************************/
void AcpiOsSleep (UINT64 Milliseconds)
{
	time_t microseconds = (uint64)Milliseconds*1000;
	thread_suspend_timeout(NULL, microseconds);
}


/******************************************************************************
 *
 * FUNCTION:     AcpiOsStall
 *
 * PARAMETERS:  Microseconds		The amount of time to delay, in microsecs.
 *
 * RETURN:      None
 *
 * DESCRIPTION: The function waits for a specified time. Execution of the 
 *				running thread is not suspended for this time. The time 
 *				granularity is one microsecond.
 *
 *****************************************************************************/
void AcpiOsStall (UINT32 Microseconds)
{
	arch_delay_usec((int64)Microseconds);
}


/******************************************************************************
 *
 * FUNCTION:     AcpiOsGetTimer
 *
 * PARAMETERS:  None
 *
 * RETURN:      TimerValue	The current value of the system timer in 
 *							100-nanoseconds units.
 *
 * DESCRIPTION: The function returns the current value of a fine-granularity
 *				64-bit system timer. This timer is used to implement the Timer 
 *				ASL/AML function.
 *
 *****************************************************************************/
UINT64 AcpiOsGetTimer (void)
{
	return (UINT64)arch_timestamp_nsec();
}


/*OSL functionallity for scheduling*/


/******************************************************************************
 *
 * FUNCTION:    init_acpi_scheduler
 *
 * PARAMETERS:  None
 *
 * RETURN:      AE_OK		The initialization of the OSL scheduling system was 
 *							successfull.
 *				AE_ERROR	The initialization of the OSL scheduling system was
 *							unsuccessfull.
 *
 * DESCRIPTION: Initializes the OSL scheduling system by:
 *				1) Initializing a request queue for storing tasks and contexts.
 *				2) Creating and starting a thread that wil retrieve and start
 *				   tasks from this queue.
 *
 *				This function should be run when initializing the OSL, for 
 *				instance during the AcpiOsInitialize function.
 *
 *****************************************************************************/
ACPI_STATUS init_acpi_scheduler(void)
{
	vxerr_t		vxerr;
	
	//Need a request queue that we can put tasks we wish to schedule on
	vxerr = _rq_new(&queueref, NULL, RESOURCE_METRIC_NONE, RQ_OWNTYPE_ANON, NULL);
	if (vxerr != VXERR_OK) {
		syslog(LOG_CRIT, "_rq_new() returned %s", strvxerr(vxerr));
		goto error;
	}

	scheduler_queue = (rqueue_t *)VxO_REFGETOBJ(&queueref);

	//Let us create the scheduler thread
    vxerr = thread_create_fmt(&tcbref, 10000, AFFINITY_NAFFINITY, KERNEL_CPUAREF, acpi_scheduler_thread, "ACPI Scheduler Thread");
	if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to create scheduling thread: %s", strvxerr(vxerr));
		goto error;
	}

	//Now, let us start it
	vxerr = thread_resume(&tcbref, 0);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to resume zero thread for cpu %d: %s", strvxerr(vxerr));
		goto error;
	}

	return AE_OK;

error:
	return AE_ERROR;
}


/******************************************************************************
 *
 * FUNCTION:    acpi_scheduler_thread
 *
 * PARAMETERS:  None
 *
 * RETURN:      None
 *
 * DESCRIPTION: A simple thread that will run forever trying to dequeue tasks 
 *				from the scheduling queue. If the thread successfully dequeues 
 *				a tasks, it will start it with the passed context.
 *
 *****************************************************************************/
static void acpi_scheduler_thread(void)
{
	objref_t    reqreflist[1];
	int64		batching_factor;
	int64		numev;
	request_t	*req;

	batching_factor = 1;
    while (1) {
        //Try to dequeue
		numev = _rq_dequeue(NULL, NULL, &queueref,  batching_factor, reqreflist, TIME_NTIME);
        if (numev == 0) {
            continue;
		} else {

			//Get the request from the list of request references
			req = (request_t *)VxO_REFGETOBJ(&reqreflist[0]);

			//Invoke the closure - in effect scheduling the task
			closure_invoke(req->req_closure);

			//Free the memory occupied by the request
			VxO_DECREF(req);
		}
    }
}


/******************************************************************************
 *
 * FUNCTION:    execution_type_is_ok
 *
 * PARAMETERS:  type	- An execution type to be tested for validity.
 *
 * RETURN:      None
 *
 * DESCRIPTION: A simple thread that will run forever trying to dequeue tasks 
 *				from the scheduling queue. If the thread successfully dequeues 
 *				a tasks, it will start it with the passed context.
 *
 *****************************************************************************/
static int execution_type_is_ok(ACPI_EXECUTE_TYPE type)
{
	switch(type)
	{
		case OSL_GLOBAL_LOCK_HANDLER:
			return 1;
		case OSL_NOTIFY_HANDLER:
			return 1;
		case OSL_GPE_HANDLER:
			return 1;
		case OSL_DEBUGGER_THREAD:
			return 1;
		case OSL_EC_POLL_HANDLER:
			return 1;
		case OSL_EC_BURST_HANDLER:
			return 1;
		default:
			return 0;
	}
}