#include <vortex/acpi/acpiheaders.h>

static int units_is_availabe(acpi_osd_semaphore_t *sem, UINT32 units);

#define SPINLOCK_LOCKED = 1
#define SPINLOCK_UNLOCKED = 0
struct acpi_osd_semaphore{
	UINT32	as_maxunits;
	UINT32	as_units;
    tcb_t	*as_queue;              ///< Threads blocked on lock
	uint64	as_waiters;
	int		as_reset;
};

type_t acpiSemaphore_type = {
    .t_name = "acpiSemaphore",
	.t_size = sizeof(acpi_osd_semaphore_t),
    .t_alloc = KOBJ_ZEROFILL,
};

type_t acpiSpinlock_type = {
    .t_name = "acpiSpinlock",
	.t_size = sizeof(acpi_osd_spinlock_t),
    .t_alloc = KOBJ_ZEROFILL,
};


/*
 * Spinlock primitives: For a thorough discussion on the why's and how's of 
 * implementing this in Vortex, see the shitwefoundout article: 
 * http://shitwefoundout.com/wiki/Spinlocks_in_Vortex 
 *
 */


#ifndef AcpiOsCreateLock
/******************************************************************************
 *
 * FUNCTION:    AcpiOsCreateLock
 *
 * PARAMETERS:  OutHandle           - A pointer to a location where a handle to 
 *									  the lock is to be returned.
 *
 * RETURN:      AE_OK				The spinlock was successfully created.
 *				AE_BAD_PARAMETER	The OutHandle is NULL.
 *				AE_NO_MEMORY		Insufficient memory to create the spinlock.
 *
 * DESCRIPTION: Create a spinlock. Spinlocks are used in the ACPICA subsystem 
 *				only when there is requirement for mutual exclusion on data 
 *				structures thate are accessed by both interrupt handlers and 
 *				normal code.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsCreateLock (ACPI_SPINLOCK *OutHandle)
{
	acpi_osd_spinlock_t *spinlock;

	if (OutHandle == NULL) {
		syslog(LOG_CRIT, "When creating spinloc, outhandle == NULL");
		return AE_BAD_PARAMETER;
	}

	spinlock = (acpi_osd_spinlock_t *)VxO_NEW(acpiSpinlock);

	if (spinlock == NULL) {
		syslog(LOG_CRIT, "Not enough memory when creating spinlock");
		return AE_NO_MEMORY;
	}

	*OutHandle = spinlock;

	return AE_OK;
}
#endif

/******************************************************************************
 *
 * FUNCTION:    AcpiOsDeleteLock
 *
 * PARAMETERS:  Handle           - A handle to a lock object that was returned 
 *								   by a previous call to AcpiOsCreateLock.
 *
 * RETURN:      None
 *
 * DESCRIPTION: Delete a spinlock.
 *
 *****************************************************************************/
void AcpiOsDeleteLock (ACPI_SPINLOCK Handle)
{
	VxO_DECREF(Handle);
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsAcquireLock
 *
 * PARAMETERS:  Handle           - A handle to a lock object that was returned 
 *								   by a previous call to AcpiOsCreateLock.
 *
 * RETURN:      Flags		Platform-dependent CPU flags. To be used when lock
 *							is released
 *
 * DESCRIPTION: Wait for and acquire spinlock. May be called from interrupt 
 *				handlers, GPE handlers, and Fixed event handlers. Single 
 *				threaded OSL implementation should always return AE_OK.
 *
 *****************************************************************************/
ACPI_CPU_FLAGS AcpiOsAcquireLock (ACPI_SPINLOCK Handle)
{
	lock_t lock;
	VxO_LOCK(&lock, Handle);
	return lock.lo_istate;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsReleaseLock
 *
 * PARAMETERS:  Handle          A handle to a lock object that was returned 
 *								by a previous call to AcpiOsCreateLock.
 *				Flags			Flags that were returned from AcpiOsAquireLock.
 *
 * RETURN:      None
 *
 * DESCRIPTION: Release a previously acquired spinlock.
 *
 *****************************************************************************/
void AcpiOsReleaseLock (ACPI_SPINLOCK Handle, ACPI_CPU_FLAGS Flags)
{
	lock_t lock;
	lock.lo_istate = Flags;
	lock.lo_mutex = VxO_OBJMUTEX(Handle);
	VxO_UNLOCK(&lock);
}




/*
 * Semaphore primitives
 */

/******************************************************************************
 *
 * FUNCTION:    AcpiOsCreateSemaphore
 *
 * PARAMETERS:  InitialUnits        - Units to be assigned to the new semaphore
 *              OutHandle           - Where a handle will be returned
 *
 * RETURN:      AE_OK				The semaphore was succesfully created.
 *				AE_BAD_PARAMETERS	At least one of the following is true:
 *										- The initial units is invalid.
 *										- The OutHandle pointer is NULL.
 *				AE_NO_MEMORY		Insufficient memory to create semaphore.
 *
 * DESCRIPTION: Create a standard semaphore. The MaxUnits parameter allows the 
 *				semaphore to be tailored to specific uses. For example, a 
 *				MaxUnits value of one indicates that the semaphore is to be 
 *				used as a mutex. The underlying OS object used to implement this 
 *				semaphore may be different than if MaxUints is greater than 
 *				one (indicating that it will be used as a general purpose 
 *				semaphore). The ACPICA Core Subsystem creates semaphores of 
 *				both the mutex and general purpose variety.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsCreateSemaphore (UINT32 MaxUnits, UINT32 InitialUnits, ACPI_SEMAPHORE *OutHandle)
{
	acpi_osd_semaphore_t *as;
	
	if (OutHandle == NULL || MaxUnits == 0 || InitialUnits > MaxUnits)
		return AE_BAD_PARAMETER;
	
	as = (acpi_osd_semaphore_t *)VxO_NEW(acpiSemaphore);
	if ( as == NULL)
		return AE_NO_MEMORY;

	as->as_maxunits = MaxUnits;
	as->as_units = InitialUnits;
	as->as_waiters = 0; 
	as->as_reset = 0;
	
	*OutHandle = as;

	return AE_OK;
}


/******************************************************************************
 *
 * FUNCTION:    AcpiOsDeleteSemaphore
 *
 * PARAMETERS:  Handle              - Handle returned by AcpiOsCreateSemaphore
 *
 * RETURN:      AE_OK				The semaphore was successfully deleted.
 *				AE_BAD_PARAMETERS	The Handle is invalid.
 *
 * DESCRIPTION: Delete an OS semaphore
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsDeleteSemaphore (ACPI_SEMAPHORE Handle)
{ 
	lock_t      lock;
	objref_t    asref;
	acpi_osd_semaphore_t *as = (acpi_osd_semaphore_t *)Handle;

	if (as == NULL)
		return AE_BAD_PARAMETER;
	
	//Get object ref to the semaphore (so that we can verify that it have not been freed)
	VxO_REFOBJ(&asref, as);

	//Initialize the lock, and lock the semaphore
	VxO_LOCKNULL(&lock);
	if (!VxO_REFLOCK(&lock, &asref))
		goto error;

	//If anynone is waiting on the semaphore, they must all receive a signal
	if (as->as_waiters > 0) 
	{
		//We start by seting the reset flag, and then broadcast on the resquest queue
		as->as_reset = 1;
		thread_resume_queue(&(as->as_queue), VxO_LOCKGETMUTEX(&lock), 0, as->as_waiters);
	}

	//Unlock the semaphore
	VxO_UNLOCK(&lock); 

	//Cleanup
	VxO_DECREF((void *)as);

	return AE_OK;

error:
	VxO_UNLOCK(&lock);
	return AE_OK;
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsWaitSemaphore
 *
 * PARAMETERS:  Handle              - Handle returned by AcpiOsCreateSemaphore
 *              Units               - How many units to wait for
 *              Timeout             - How long the caller is willing to wait
 *									for the requested units. The timout is 
 *									specified in milliseconds. A value of 
 *									0xFFFF (-1) indicates that the calling 
 *									thread is willing to wait forevers
 *
 * RETURN:      AE_OK				The requested units were successfully
 *									received.
 *				AE_BAD_PARAMETERS	The Handle is invalid.
 *				AE_TIME				The units could not be acquired within the 
 *									specified time limit.
 *
 * DESCRIPTION: Wait for the specified  number of units from the semaphore.
 *
 * IMPLEMTATION NOTES:	1.	The implementation of this interface must support 
 *							timout values of zero. This is frequently used to 
 *							determine if a call to the interface would block.
 *							In this case, the function must return AE_OK if 
 *							the units were obtained immediately, or an AE_TIME 
 *							to indicate that the requested units are not 
 *							available. Single threaded OSL implementations 
 *							should always return AE_OK for this interface.
 *
 *						2.	The implementation must also support arbitrary 
 *							timed waits in order for ASL functions such as 
 *							Wait() to function properly.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsWaitSemaphore (ACPI_SEMAPHORE Handle, UINT32 Units, UINT16 Timeout)
{
	vxerr_t     vxerr;
	int			timeout;
	objref_t    asref;
    lock_t      lock;
	acpi_osd_semaphore_t *as = (acpi_osd_semaphore_t *)Handle;
	ACPI_STATUS status = AE_OK;

	if (as == NULL || Units == 0) {
		syslog(LOG_CRIT, "Bad params when waiting for semaphore %U. Units == %u", (uint64)Handle, (uint32)Units);
		return AE_BAD_PARAMETER;
	}

	//Get object ref to the semaphore (so that we can verify that it has not been freed)
	VxO_REFOBJ(&asref, as);

	//Initialize the lock, and lock the semaphore
	VxO_LOCKNULL(&lock);
	if (!VxO_REFLOCK(&lock, &asref))
		goto error;

	if (as->as_maxunits != ACPI_NO_UNIT_LIMIT && as->as_maxunits < Units) {
		syslog(LOG_DEBUG, "Asking for more that is supported in Semaphore %U", (uint64)Handle);
		VxO_UNLOCK(&lock);
		return AE_LIMIT;
	}

	switch(Timeout)
	{
		//In the case that we do not want to wait
		case ACPI_DO_NOT_WAIT:
			if (!units_is_availabe(as, Units))
				status = AE_TIME;
			break;
		//In the case that the caller is willing to wait forever
		case ACPI_WAIT_FOREVER:

			while (!units_is_availabe(as, Units)) {
				as->as_waiters++;
				vxerr =  thread_suspend_queue_timeout(&(as->as_queue), VxO_LOCKGETMUTEX(&lock), TIME_NTIME); 

				if (!VxO_REFCHECK(&asref))
					goto error;

				as->as_waiters--;

				if (vxerr != VXERR_OK || as->as_reset) {
					status = AE_ERROR;
					break;
				}
			}

			break;
		//The default case where we wish to wait for a given time
		default:
			//ACPICA timeouts are in ms, while Vortex operate with microseconds.
			timeout = Timeout*1000;
			
			while(!units_is_availabe(as, Units)) {
				as->as_waiters++;
				vxerr = thread_suspend_queue_timeout(&(as->as_queue), VxO_LOCKGETMUTEX(&lock), timeout); 

				// We have the semaphore, but the semaphore may have been destroyes since thread_resume_queue() drops the lock
				if (!VxO_REFCHECK(&asref))
					goto error;

				as->as_waiters--;

				if(as->as_reset == 1) {
					status = AE_ERROR;
					break;
				}

				//Units are availabe, so we can break from the loop
				if (units_is_availabe(as, Units))
					break;

				//Check whether or not we have timed out
				if (vxerr == VXERR_TIMEOUT) {
					status = AE_TIME;
					break;
				}
			}
	}

	//Take action according to whether or not the requested units were obtainable
	if (ACPI_SUCCESS(status))
		as->as_units -= Units;

	VxO_UNLOCK(&lock);
	return status;

error:
	syslog(LOG_CRIT, "Something terrible happened when attemtping to lock semaphore %U", (uint64)Handle);
    VxO_UNLOCK(&lock);
	status = AE_ERROR;
	return status;
}

/******************************************************************************
 *
 * FUNCTION:    AcpiOsSignalSemaphore
 *
 * PARAMETERS:  Handle              - Handle returned by AcpiOsCreateSemaphore
 *              Units               - Number of units to send
 *
 * RETURN:      AE_OK				The semaphore was successfully signaled.
 *				AE_BAD_PARAMETERS	The handle is invalid.
 *				AE_LIMIT			The semaphore has already been signaled 
 *									MaxUnits times. No more units can be 
 *									accepted.
 *
 * DESCRIPTION: Send the requested number of units to a semaphore. Single 
 *				threaded OSL implementations should always return AE_OK
 *				for this interface.
 *
 *****************************************************************************/
ACPI_STATUS AcpiOsSignalSemaphore (ACPI_SEMAPHORE Handle, UINT32 Units)
{
	vxerr_t     vxerr;
	objref_t    asref;
    lock_t      lock;
	acpi_osd_semaphore_t *as = (acpi_osd_semaphore_t *)Handle;


	if (as == NULL || Units == 0) {
		syslog(LOG_CRIT, "Bad params when signaling Semaphore %U", (uint64)Handle);
		return AE_BAD_PARAMETER;
	}

	//Get object ref to the semaphore (so that we can verify that it have not been freed)
	VxO_REFOBJ(&asref, as);

	//Initialize the lock, and lock the semaphore
	VxO_LOCKNULL(&lock);
	if (!VxO_REFLOCK(&lock, &asref))
		goto error;

	if (as->as_maxunits != ACPI_NO_UNIT_LIMIT) {
		if (as->as_maxunits < Units || as->as_maxunits - Units < as->as_units) {
			VxO_UNLOCK(&lock);
			return AE_LIMIT;
		}
	}

	as->as_units += Units;
	if (as->as_waiters > 0 && units_is_availabe(as, Units)) {
		/*This is a bit sneaky. Since we know that each of the blocked threads have asked for 
		 * at least 1 unit, we can avoid the additional overhead incurred by broadcasting if we 
		 * resume only the maximum number of threads that possibly can be resumed (in the case 
		 * that they each just wait for one unit*/
		vxerr = thread_resume_queue(&(as->as_queue), VxO_LOCKGETMUTEX(&lock), 0, (int64)Units);
        if (vxerr != VXERR_OK) {
            syslog(LOG_CRIT, "Unable to resume thread...");
            goto error;
        }
	}

	VxO_UNLOCK(&lock);
	return AE_OK;

error:
	syslog(LOG_CRIT, "Something horrible happened when signaling Semaphore %U", (uint64)Handle);
	VxO_UNLOCK(&lock);
	return AE_OK;
}



/*Helper functions to keep things a bit more tidy*/
static int units_is_availabe(acpi_osd_semaphore_t *semaphore, UINT32 units)
{
	if (semaphore->as_units >= units)
		return 1;
	else
		return 0;
}



/*
 * Mutex primitives. May be configured to use semaphores instead via
 * ACPI_MUTEX_TYPE (see platform/acenv.h)
 */
#if (ACPI_MUTEX_TYPE != ACPI_BINARY_SEMAPHORE)

ACPI_STATUS AcpiOsCreateMutex (ACPI_MUTEX *OutHandle)
{
	syslog(LOG_ERROR, "Call to AcpiOsCreateMutex - should have been disabled\n");
}

void AcpiOsDeleteMutex (ACPI_MUTEX Handle)
{
	syslog(LOG_ERROR, "Call to AcpiOsDeleteMutex - should have been disabled\n");
}

ACPI_STATUS AcpiOsAcquireMutex (ACPI_MUTEX Handle, UINT16 Timeout)
{
	syslog(LOG_ERROR, "Call to AcpiOsAcquireMutex - should have been disabled\n");
}

void AcpiOsReleaseMutex (ACPI_MUTEX Handle)
{
	syslog(LOG_ERROR, "Call to AcpiOsReleaseMutex - should have been disabled\n");
}
#endif