#include <vortex/acpi/acpiheaders.h>
#include <vortex/acpi/platform/acpi_vortex.h>
#include "../core/arch/x86_64/cpu.h"
#include <vortex/cpumux.h>

#define VORTEX_ACPI_DEBUG 1


typedef struct _AcpiSysInfo {
	UINT32 AcpiCaVersion;
	UINT32 Flags;
	UINT32 TimerResoulution;
	UINT32 Reserved1;
	UINT32 Reserved2;
	UINT32 DebugLevel;
	UINT32 DebugLayer;
} ;

struct pm_tree_node {
	char			path[256];
	pm_tree_node_t	*parent, *next, *prev, *children;
	device_t		*device;
};

struct pm_tree {
	pm_tree_node_t	*root;
	uint32_t		depth;
};

type_t pmtreenode_type = {
    .t_name = "pmtreenode",
	.t_size = sizeof(pm_tree_node_t),
    .t_alloc = KOBJ_ZEROFILL,
};


ACPI_OBJECT *execute_acpi_object_with_return_type_package(const utf8_t *object_path);

//Forward declaration of convenience functions for working with C-states
static int32 get_num_cstates(ACPI_OBJECT *cst);
//static ACPI_OBJECT *execute_CST(const utf8_t *object_path);
static bool_t CST_is_valid(ACPI_OBJECT *cst);

void dfs_pm_tree(pm_tree_t *tree)
{
	pm_tree_node_t	*current;
	dict_t			*visited_children;
	vxerr_t			vxerr;
	objref_t		childrenref;

	/* Start by creating a dict for visited children. This does 
	 * not waste as much memory as one might think, asall the 
	 * children of a node is linked together in a list.*/
	visited_children = VxD_NEW(SYNC_GLOBAL, DICT_TYPE_IKEY);
	if (visited_children == NULL)
		syslog(LOG_CRIT, "Unable to create visited children dict");

	/*Start the search by going to the root children*/
	VxO_REFOBJ(&childrenref, tree->root->children);
	vxerr = VxD_SETITEM(&childrenref, visited_children, (uid_t)childrenref.ob_uid, NULL);
	current = tree->root->children;
	if (vxerr != VXERR_OK)
		syslog(LOG_CRIT, "Unable to add node to children dict");

	while (current != NULL) {
		/*Do what we want to do, for instance check a name*/
		syslog(LOG_DEBUG, "Now visiting %s", current->path);

		/*If this is the root node, we are done*/
		if (current->parent == NULL)
			break;

		/*If we have children, get a reference to them and check if they have been visited*/
		if (current->children != NULL) {
			VxO_REFOBJ(&childrenref, current->children);
			vxerr = VxD_GETITEM(&childrenref, visited_children, (uid_t)childrenref.ob_uid, NULL);
		}

		/*If we have children and these are unvisited - visit them*/
		if (current->children != NULL && vxerr == VXERR_NEXIST) {

			VxO_REFOBJ(&childrenref, current->children);
			vxerr = VxD_SETITEM(&childrenref, visited_children, (uid_t)childrenref.ob_uid, NULL);
			current = current->children;
			if (vxerr != VXERR_OK)
				syslog(LOG_CRIT, "Unable to add node to children dict");
			
		}
		else if (current->next != NULL)
			current = current->next;
		else
			current = current->parent;
	}

	syslog(LOG_DEBUG, "Done searching");
}


ACPI_STATUS AcpiPMNsWalkNamespace (ACPI_OBJECT_TYPE Type, ACPI_HANDLE StartNode, UINT32 MaxDepth, UINT32 Flags, ACPI_WALK_CALLBACK DescendingCallback, ACPI_WALK_CALLBACK AscendingCallback, void **Context, void **ReturnValue)
{
    ACPI_STATUS             Status;
    ACPI_STATUS             MutexStatus;
    ACPI_NAMESPACE_NODE     *ChildNode;
    ACPI_NAMESPACE_NODE     *ParentNode;
    ACPI_OBJECT_TYPE        ChildType;
    UINT32                  Level;
    BOOLEAN                 NodePreviouslyVisited = FALSE;
	pm_tree_node_t			*current;

    /* Special case for the namespace Root Node */
    if (StartNode == ACPI_ROOT_OBJECT) {
        StartNode = AcpiGbl_RootNode;

		/*Create a root node for our tree*/
		current = (pm_tree_node_t *)VxO_NEW(pmtreenode);
		AcpiUtStrcpy(current->path, "Root Node");
		Context = (void **)&current;
    }

    /* Null child means "get first node" */
    ParentNode  = StartNode;
    ChildNode   = AcpiNsGetNextNode (ParentNode, NULL);
    ChildType   = ACPI_TYPE_ANY;
    Level       = 1;

    /*
     * Traverse the tree of nodes until we bubble back up to where we
     * started. When Level is zero, the loop is done because we have
     * bubbled up to (and passed) the original parent handle (StartEntry)
     */
    while (Level > 0 && ChildNode) {
        Status = AE_OK;

        /* Found next child, get the type if we are not searching for ANY */
        if (Type != ACPI_TYPE_ANY)
            ChildType = ChildNode->Type;
        
        /*
         * Ignore all temporary namespace nodes (created during control
         * method execution) unless told otherwise. These temporary nodes
         * can cause a race condition because they can be deleted during
         * the execution of the user function (if the namespace is
         * unlocked before invocation of the user function.) Only the
         * debugger namespace dump will examine the temporary nodes.
         */
        if ((ChildNode->Flags & ANOBJ_TEMPORARY) && !(Flags & ACPI_NS_WALK_TEMP_NODES))
            Status = AE_CTRL_DEPTH;

        /* Type must match requested type */
        else if (ChildType == Type) {
            /*
             * Found a matching node, invoke the user callback function.
             * Unlock the namespace if flag is set.
             */
            if (Flags & ACPI_NS_WALK_UNLOCK) {
                MutexStatus = AcpiUtReleaseMutex (ACPI_MTX_NAMESPACE);
                if (ACPI_FAILURE (MutexStatus))
                    return_ACPI_STATUS (MutexStatus);
            }

            /*
             * Invoke the user function, either descending, ascending,
             * or both.
             */
            if (!NodePreviouslyVisited) {
                if (DescendingCallback)
                    Status = DescendingCallback (ChildNode, Level, Context, ReturnValue);
            } else {
                if (AscendingCallback) {
                    Status = AscendingCallback (ChildNode, Level,
                                Context, ReturnValue);
                }
            }

            if (Flags & ACPI_NS_WALK_UNLOCK) {
                MutexStatus = AcpiUtAcquireMutex (ACPI_MTX_NAMESPACE);
                if (ACPI_FAILURE (MutexStatus))
                    return_ACPI_STATUS (MutexStatus);
            }
			
            switch (Status)
            {

            case AE_OK:
            case AE_CTRL_DEPTH:
                // Just keep going //
                break;

            case AE_CTRL_TERMINATE:
                // Exit now, with OK status //
                return_ACPI_STATUS (AE_OK);

            default:
                // All others are valid exceptions //
                return_ACPI_STATUS (Status);
            }
			
        }

        /*
         * Depth first search: Attempt to go down another level in the
         * namespace if we are allowed to. Don't go any further if we have
         * reached the caller specified maximum depth or if the user
         * function has specified that the maximum depth has been reached.
         */
        if (!NodePreviouslyVisited && (Level < MaxDepth) && (Status != AE_CTRL_DEPTH)) {
            if (ChildNode->Child) {
                /* There is at least one child of this node, visit it */
                Level++;
                ParentNode = ChildNode;
                ChildNode = AcpiNsGetNextNode (ParentNode, NULL);
                continue;
            }
        }

        /* No more children, re-visit this node */
        if (!NodePreviouslyVisited) {
            NodePreviouslyVisited = TRUE;
            continue;
        }

        /* No more children, visit peers */
        ChildNode = AcpiNsGetNextNode (ParentNode, ChildNode);
        if (ChildNode)
            NodePreviouslyVisited = FALSE;

        /* No peers, re-visit parent */
        else {
            /*
             * No more children of this node (AcpiNsGetNextNode failed), go
             * back upwards in the namespace tree to the node's parent.
             */
            Level--;
            ChildNode = ParentNode;
            ParentNode = ParentNode->Parent;

            NodePreviouslyVisited = TRUE;
			/*Remember to go one step up in the tree being created as well*/
			current = (pm_tree_node_t *)*Context;
			if (current->parent != NULL) {
				syslog(LOG_DEBUG, "    Before changing context: %s", current->path);
				current = current->parent;
				syslog(LOG_DEBUG, "    After changing context: %s", current->path);
				*Context = (void *)current;
			}
			else
				continue;

        }
    }
	current = (pm_tree_node_t *)*Context;
	syslog(LOG_DEBUG, "Before leaving, the path of thecurrent node of the context is: %s", current->path);
	*ReturnValue = current;
	
    /* Complete walk, not terminated by user function */
    return_ACPI_STATUS (AE_OK);
}


ACPI_STATUS
AcpiPMWalkNamespace (
    ACPI_OBJECT_TYPE        Type,
    ACPI_HANDLE             StartObject,
    UINT32                  MaxDepth,
    ACPI_WALK_CALLBACK      DescendingCallback,
    ACPI_WALK_CALLBACK      AscendingCallback,
    void                    **Context,
    void                    **ReturnValue)
{
    ACPI_STATUS             Status;

	pm_tree_node_t *context_node;
    ACPI_FUNCTION_TRACE (AcpiWalkNamespace);


    /* Parameter validation */
    if ((Type > ACPI_TYPE_LOCAL_MAX) ||
        (!MaxDepth)                  ||
        (!DescendingCallback && !AscendingCallback))
    {
        return_ACPI_STATUS (AE_BAD_PARAMETER);
    }

    /*
     * Need to acquire the namespace reader lock to prevent interference
     * with any concurrent table unloads (which causes the deletion of
     * namespace objects). We cannot allow the deletion of a namespace node
     * while the user function is using it. The exception to this are the
     * nodes created and deleted during control method execution -- these
     * nodes are marked as temporary nodes and are ignored by the namespace
     * walk. Thus, control methods can be executed while holding the
     * namespace deletion lock (and the user function can execute control
     * methods.)
     */
    Status = AcpiUtAcquireReadLock (&AcpiGbl_NamespaceRwLock);
    if (ACPI_FAILURE (Status))
        return_ACPI_STATUS (Status);

    /*
     * Lock the namespace around the walk. The namespace will be
     * unlocked/locked around each call to the user function - since the user
     * function must be allowed to make ACPICA calls itself (for example, it
     * will typically execute control methods during device enumeration.)
     */
    Status = AcpiUtAcquireMutex (ACPI_MTX_NAMESPACE);
    if (ACPI_FAILURE (Status))
        goto UnlockAndExit;

    /* Now we can validate the starting node */
    if (!AcpiNsValidateHandle (StartObject))
        Status = AE_BAD_PARAMETER;
        goto UnlockAndExit2;


    Status = AcpiPMNsWalkNamespace (Type, StartObject, MaxDepth, ACPI_NS_WALK_UNLOCK, DescendingCallback, AscendingCallback, Context, ReturnValue);
	context_node = (pm_tree_node_t *)*ReturnValue;
	syslog(LOG_DEBUG, "context_node's path is: %s", context_node->path);

UnlockAndExit2:
    (void) AcpiUtReleaseMutex (ACPI_MTX_NAMESPACE);

UnlockAndExit:
    (void) AcpiUtReleaseReadLock (&AcpiGbl_NamespaceRwLock);
    return_ACPI_STATUS (Status);
}


/*Functions for displaying information - mostly for testing purposes*/
ACPI_STATUS display_processor_devices(void)
{
	ACPI_HANDLE		ProcessorsHandle;	
	AcpiGetHandle(ACPI_ROOT_OBJECT, "_PR_", &ProcessorsHandle);
	syslog(LOG_DEBUG, "Display all of the Processor namespace info ");
	AcpiWalkNamespace( ACPI_TYPE_ANY, ProcessorsHandle, INT32_MAX, (ACPI_WALK_CALLBACK)DisplayProcessorDevice, (ACPI_WALK_CALLBACK)NULL, NULL, (void**)NULL);
}

void display_all_devices(void)
{
	syslog(LOG_DEBUG, "Display all of the devices in the namespace");
	AcpiWalkNamespace(ACPI_TYPE_ANY, ACPI_ROOT_OBJECT, INT32_MAX, (ACPI_WALK_CALLBACK)DisplayOneDevice, (ACPI_WALK_CALLBACK)NULL, NULL, (void**)NULL);
}





void build_pm_device_tree(void)
{
	pm_tree_t tree;
	pm_tree_node_t *new_root;
	tree.depth = 0; 
	AcpiPMWalkNamespace(ACPI_TYPE_ANY, ACPI_ROOT_OBJECT, INT32_MAX, (ACPI_WALK_CALLBACK)evaluate_node, (ACPI_WALK_CALLBACK)NULL, (void **)&(tree.root),(void **)&new_root);
	syslog(LOG_DEBUG, "New root path is: %s", new_root->path);
	tree.root = new_root;
	dfs_pm_tree(&tree);
}

void *evaluate_node(ACPI_HANDLE ObjHandle, UINT32 Level, void **Context)
{
	ACPI_STATUS			Status;
	ACPI_DEVICE_INFO	*Info;
	ACPI_BUFFER			Path;
	char				Buffer[256];
	pm_tree_node_t		*parent, *current;
	ACPI_HANDLE			AddressHandle;
	ACPI_BUFFER         ReturnValue;

	ReturnValue.Length = ACPI_ALLOCATE_BUFFER;
	Path.Length = sizeof(Buffer);
	Path.Pointer = Buffer;

	/*Get the full path of the device and print it*/
	Status = AcpiNsHandleToPathname(ObjHandle, &Path);
	if (ACPI_SUCCESS(Status))
		syslog(LOG_DEBUG, "%s", Path.Pointer);

	/*Get device info and print it*/
	Status = AcpiGetObjectInfo(ObjHandle, &Info);
	if (ACPI_SUCCESS(Status)) {
		if (Info->Type == ACPI_TYPE_DEVICE || Info->Type == ACPI_TYPE_PROCESSOR) {
			parent = (pm_tree_node_t *)*Context;
			current = (pm_tree_node_t *)VxO_NEW(pmtreenode);
			AcpiUtStrcpy(current->path, (const char *)(Path.Pointer));
			current->parent = parent;
			VxL_ENQUEUE(parent->children, current);

			if (Info->Type == ACPI_TYPE_DEVICE) {
				/*Check if we have a child _ADR object, and if so evaluate it*/
				Status = AcpiGetHandle(ObjHandle, "_ADR", &AddressHandle);
				if (Status == AE_OK) {
					Status = AcpiEvaluateObject (AddressHandle, NULL,  NULL, &ReturnValue);
					syslog(LOG_DEBUG, "After evaluating the _ADR object, status is : %d", Status);
					syslog(LOG_DEBUG, "The returned _ADR was: %x", *(uint32 *)&(ReturnValue.Pointer));
				}
			}

			syslog(LOG_DEBUG, "Inserted %s into the children list of %s - Now updating Context to match this new node", (const char *)(Path.Pointer), parent->path);
			*Context = (void *)current;			
		}

	}
	return NULL;
}

ACPI_STATUS print_acpi_system_info(void)
{
	ACPI_STATUS status;
	ACPI_BUFFER acpi_buffer;
	acpi_sysinfo_t sysinfo;
	acpi_buffer.Pointer = (void *)&sysinfo;
	acpi_buffer.Length = (UINT64)sizeof(acpi_sysinfo_t);

	syslog(LOG_DEBUG, "#############################################");
	syslog(LOG_DEBUG, "########## ACPI SYSTEM INFORMATION ##########");
	syslog(LOG_DEBUG, "#############################################");

	status = AcpiGetSystemInfo(&acpi_buffer);

	switch(status)
	{
	case AE_OK:
		syslog(LOG_DEBUG, "Got OK from AcpiGetSystemInfo");
		
		syslog(LOG_DEBUG, "AcpiCaVersion: %x", sysinfo.AcpiCaVersion);

		if (sysinfo.Flags == ACPI_SYS_MODE_ACPI)
			syslog(LOG_DEBUG, "ACPI mode is supported on this system");
		else if (sysinfo.Flags == ACPI_SYS_MODE_LEGACY)
			syslog(LOG_DEBUG, "Legacy mode is supported on this system");


		syslog(LOG_DEBUG, "Timer Resolution: %u", (uint32)sysinfo.TimerResoulution);
		syslog(LOG_DEBUG, "Debug Level: %u", (uint32)sysinfo.DebugLevel);
		syslog(LOG_DEBUG, "Debug Layer: %u", (uint32)sysinfo.DebugLayer);

		break;
	case AE_BAD_PARAMETER:
		syslog(LOG_CRIT, "Got bad params from AcpiGetSystemInfo");
		return AE_BAD_PARAMETER;
	case AE_BUFFER_OVERFLOW:
		syslog(LOG_CRIT, "Got buffer overflow from AcpiGetSystemInfo");
		return AE_BUFFER_OVERFLOW;
	}

	syslog(LOG_DEBUG, "#############################################");
	return AE_OK;
}

void print_fadt_info(void)
{
	syslog(LOG_DEBUG, "#############################################");
	syslog(LOG_DEBUG, "########### ACPI FADT INFORMATION ###########");
	syslog(LOG_DEBUG, "#############################################");
	syslog(LOG_DEBUG, " ");
	syslog(LOG_DEBUG, "Duty Offset: %x", AcpiGbl_FADT.DutyOffset);
	syslog(LOG_DEBUG, "Duty Width: %x", AcpiGbl_FADT.DutyWidth);
	syslog(LOG_DEBUG, "Smi Command: %x", AcpiGbl_FADT.SmiCommand);
	syslog(LOG_DEBUG, "Pstate Control: %x", AcpiGbl_FADT.PstateControl);
	syslog(LOG_DEBUG, "C2 Worst case latency: %u", AcpiGbl_FADT.C2Latency);
	syslog(LOG_DEBUG, "C3 Worst case latency: %u", AcpiGbl_FADT.C3Latency);
	syslog(LOG_DEBUG, "OEM: %s", AcpiGbl_FADT.Header.OemId);
	syslog(LOG_DEBUG, "#############################################");
}

void print_ACPI_error_status(ACPI_STATUS Status)
{
	if (Status == AE_BAD_PARAMETER)
		syslog(LOG_DEBUG, "AE_BAD_PARAMETER");
	if (Status == AE_LIMIT)
		syslog(LOG_DEBUG, "AE_LIMIT");
	if (Status == AE_AML_BAD_OPCODE)
		syslog(LOG_DEBUG, "AE_BAD_OPCODE");
	if (Status == AE_AML_NO_OPERAND)
		syslog(LOG_DEBUG, "AE_AML_NO_OPERAND");
	if (Status == AE_AML_OPERAND_TYPE)
		syslog(LOG_DEBUG, "AE_AML_OPERAND_TYPE");
	if (Status == AE_AML_OPERAND_VALUE)
		syslog(LOG_DEBUG, "AE_AML_OPERAND_VALUE");
	if (Status == AE_AML_UNINITIALIZED_LOCAL)
		syslog(LOG_DEBUG, "AE_AML_UNINITIALIZED_LOCAL");
	if (Status == AE_AML_UNINITIALIZED_ELEMENT)
		syslog(LOG_DEBUG, "AE_AML_UNINITIALIZED_ELEMENT");
	if (Status == AE_AML_NUMERIC_OVERFLOW)
		syslog(LOG_DEBUG, "AE_NUMERIC_OVERFLOW");
	if (Status == AE_AML_REGION_LIMIT)
		syslog(LOG_DEBUG, "AE_AML_REGION_LIMIT");
	if (Status == AE_AML_BUFFER_LIMIT)
		syslog(LOG_DEBUG, "AE_AML_BUFFER_LIMIT");
	if (Status == AE_AML_PACKAGE_LIMIT)
		syslog(LOG_DEBUG, "AE_AML_PACKAGE_LIMIT");
	if (Status == AE_AML_DIVIDE_BY_ZERO)
		syslog(LOG_DEBUG, "AE_AML_DIVIDE_BY_ZERO");
	if (Status == AE_AML_BAD_NAME)
		syslog(LOG_DEBUG, "AE_AML_BAD_NAME");
	if (Status == AE_AML_NAME_NOT_FOUND)
		syslog(LOG_DEBUG, "AE_AML_NAME_NOT_FOUND");
	if (Status == AE_AML_INTERNAL)
		syslog(LOG_DEBUG, "AE_AML_INTERNAL");
	if (Status == AE_BAD_CHARACTER)
		syslog(LOG_DEBUG, "AE_BAD_CHARACTER");
	if (Status == AE_BAD_DATA)
		syslog(LOG_DEBUG, "AE_BAD_DATA");
	if (Status == AE_BAD_PATHNAME)
		syslog(LOG_DEBUG, "AE_BAD_PATHNAME");
	if (Status == AE_BUFFER_OVERFLOW)
		syslog(LOG_DEBUG, "AE_BUFFER_OVERFLOW");
	if (Status == AE_ERROR)
		syslog(LOG_DEBUG, "AE_ERROR");
	if (Status == AE_NO_MEMORY)
		syslog(LOG_DEBUG, "AE_NO_MEMORY");
	if (Status == AE_NOT_FOUND)
		syslog(LOG_DEBUG, "AE_NOT_FOUND");
	if (Status == AE_NULL_OBJECT)
		syslog(LOG_DEBUG, "AE_NULL_OBJECT");
	if (Status == AE_STACK_OVERFLOW)
		syslog(LOG_DEBUG, "AE_STACK_OVERFLOW");
	if (Status == AE_STACK_UNDERFLOW)
		syslog(LOG_DEBUG, "AE_STACK_UNDEFLOW");
	if (Status == AE_TYPE)
		syslog(LOG_DEBUG, "AE_TYPE");
}

void print_device_status(uint32 status_bitmap)
{
	uint32 present		= 0x00000001;
	uint32 enabled		= 0x00000002;
	uint32 functioning	= 0x00000008;

	if (status_bitmap & present) {
		syslog(LOG_DEBUG, "Device is present");

		if (status_bitmap & enabled)
			syslog(LOG_DEBUG, "Device is enabled");
		else
			syslog(LOG_DEBUG, "Device not enabled");

		if (status_bitmap & functioning)
			syslog(LOG_DEBUG, "Device is functioning");
		else
			syslog(LOG_DEBUG, "Device not functioning");
	} else
		syslog(LOG_DEBUG, "Device not present");
}

void print_device_name(uint32 name_bitfield)
{
	char				name[5];
	char				*name_pointer;
	int i;

	name[4] = '\0';
	name_pointer = (char *)&(name_bitfield);
	for (i = 0 ; i < 4 ; i++)
	{
		name[i] = *name_pointer++;
	}
	syslog(LOG_DEBUG, "Device Name: %s", name);
}



/*Callback functions*/
void *DisplayOneDevice(ACPI_HANDLE ObjHandle, UINT32 Level, void *Context)
{
	ACPI_STATUS			Status;
	ACPI_DEVICE_INFO	*Info;
	ACPI_BUFFER			Path;
	ACPI_PNP_DEVICE_ID	*HardwareId;
	char				Buffer[256];

	Path.Length = sizeof(Buffer);
	Path.Pointer = Buffer;

	/*Get the full path of the device and print it*/
	Status = AcpiNsHandleToPathname(ObjHandle, &Path);
	if (ACPI_SUCCESS(Status)) {

		syslog(LOG_DEBUG, "######################################");
		syslog(LOG_DEBUG, "%s", Path.Pointer);
	}
	
	/*Get device info and print it*/
	Status = AcpiGetObjectInfo(ObjHandle, &Info);
	if (ACPI_SUCCESS(Status)) {
	
		if (Info->Type == ACPI_TYPE_DEVICE) {
			HardwareId = &(Info->HardwareId);
			if (Info->Valid & ACPI_VALID_STA) {
				syslog(LOG_DEBUG, "Plug and Play Unique ID: %d", Info->UniqueId);
				print_device_name(Info->Name);
				if (HardwareId->String != NULL)
					syslog(LOG_DEBUG, "Plug and Play HID: %s", HardwareId->String);
				else
				{
				
				}



			} else
				syslog(LOG_DEBUG, "_STA not found valid");
		}
		else if ( Info->Type == ACPI_TYPE_METHOD) {
			syslog(LOG_DEBUG, "This is a method");
			print_device_name(Info->Name);
		} else if ( Info->Type == ACPI_TYPE_POWER) {
			syslog(LOG_DEBUG, "This is a power object");
		} else if (Info->Type == ACPI_TYPE_BUFFER) {
			syslog(LOG_DEBUG, "This is a buffer object");
		} else if (Info->Type == ACPI_TYPE_INTEGER) {
			syslog(LOG_DEBUG, "This is an integer object");
		} else if (Info->Type == ACPI_TYPE_PACKAGE) {
			syslog(LOG_DEBUG, "This is a package object");
		} else if (Info->Type == ACPI_TYPE_STRING) {
			syslog(LOG_DEBUG, "This is a string object");
		}

	}
	return NULL;
}


/*Functions for executing queries/ working with ACPI objects*/
void ExecuteCPUOSCQueries (void)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_LIST        ArgList;
    ACPI_OBJECT             Args[4];
    ACPI_BUFFER             ReturnValue;
    ACPI_OBJECT             *Object;
	UINT8					UUID[16] = {0x16, 0xA6, 0x77, 0x40, 0x0C, 0x29, 0xBE, 0x47, 0x9E, 0xBD, 0xD8, 0x70, 0x58, 0x71, 0x39, 0x53};
	UINT32					Capabilities[2] = {_OSC_DWORD1_QUERY, 0x000006FC};
	UINT32					*ReturnedWords;
	const utf8_t			*OSCPaths[8] = {"\\_PR_.CPU0._OSC", "\\_PR_.CPU1._OSC", "\\_PR_.CPU2._OSC", "\\_PR_.CPU3._OSC", "\\_PR_.CPU4._OSC", "\\_PR_.CPU5._OSC", "\\_PR_.CPU6._OSC", "\\_PR_.CPU7._OSC"};
	UINT32					numCPUs, i;
	numCPUs = 8;


    /* Setup input arguments */
    ArgList.Count = 4;
    ArgList.Pointer = Args;

	Args[0].Type = ACPI_TYPE_BUFFER;
	Args[0].Buffer.Pointer = UUID;
	Args[0].Buffer.Length = sizeof(UUID);

	Args[1].Type = ACPI_TYPE_INTEGER;
	Args[1].Integer.Value = 1;
	
	Args[2].Type = ACPI_TYPE_INTEGER;
	Args[2].Integer.Value = 2; //DWORD defined per intel manual to be 32 bits

	Args[3].Type = ACPI_TYPE_BUFFER;
	Args[3].Buffer.Pointer = (uint8 *)Capabilities;
	Args[3].Buffer.Length = sizeof(Capabilities);


    /* Ask ACPICA to allocate space for the return object */
    ReturnValue.Length = ACPI_ALLOCATE_BUFFER;

	for (i = 0; i < numCPUs; i++) {
	
		Status = AcpiEvaluateObject (NULL, (ACPI_STRING)OSCPaths[i], &ArgList, &ReturnValue);
		if (ACPI_FAILURE (Status)) {
			print_ACPI_error_status(Status);
			syslog(LOG_CRIT, "Error while executing %s", OSCPaths[i]);
			return;
		}

		/* Ensure that the return object is large enough */
		if (ReturnValue.Length < sizeof (ACPI_OBJECT)) {
			syslog(LOG_DEBUG, "Return value from %s method too small, %.8X\n", OSCPaths[i], ReturnValue.Length);
			return;
		}

		/* Expect a buffer return value from execution of _OSI */
		Object = (ACPI_OBJECT *)ReturnValue.Pointer;
		if (Object->Type != ACPI_TYPE_BUFFER)
			syslog(LOG_DEBUG, "Invalid return type from %s, %.2X\n", OSCPaths[i], Object->Type);

		ReturnedWords = (uint32 *)(Object->Buffer.Pointer);
		syslog(LOG_DEBUG, "%s returned Dword1: %x, Dword2: %x", OSCPaths[i], ReturnedWords[0], ReturnedWords[1]);

	}

	AcpiOsFree (Object);
    return;  
}

int execute_PPC(const utf8_t *object_path)
{
	ACPI_STATUS             Status;
	ACPI_BUFFER             ReturnValue;
	ACPI_OBJECT             Object;
	UINT32					ReturnedInt;
	

	ReturnValue.Length = sizeof(Object);
	ReturnValue.Pointer = &Object;

	Status = AcpiEvaluateObject (NULL, (ACPI_STRING)object_path, NULL, &ReturnValue);
	if (ACPI_FAILURE (Status)) {
		print_ACPI_error_status(Status);
		syslog(LOG_CRIT, "Error while executing %s", object_path);
		return -1;
	}

	/* Ensure that the return object is large enough */
	if (ReturnValue.Length < sizeof (ACPI_OBJECT)) {
		syslog(LOG_DEBUG, "Return value from %s method too small, %.8X\n", object_path, ReturnValue.Length);
		return -1;
	}

	/* Expect an integer return value from execution of _PPC */
	if (Object.Type != ACPI_TYPE_INTEGER) {
		syslog(LOG_DEBUG, "Invalid return type from %s, %.2X\n", object_path, Object.Type);
		return -1;
	}

	/* Get the return value from the returned object */
	ReturnedInt = (UINT32)(Object.Integer.Value);
	return ReturnedInt;
}

ACPI_OBJECT *execute_acpi_object_with_return_type_package(const utf8_t *object_path)
{
	ACPI_STATUS		Status;
	ACPI_BUFFER		ReturnValue;
	ACPI_OBJECT		*Object;

	if (object_path == NULL) {
		syslog(LOG_DEBUG, "ERROR: In execute_acpi_object_with_return_type_package - object_path pointer is NULL");
		return NULL;
	}

	ReturnValue.Length = ACPI_ALLOCATE_BUFFER;
	Status = AcpiEvaluateObject (NULL, (ACPI_STRING)object_path, NULL, &ReturnValue);
	if (ACPI_FAILURE (Status)) {
		print_ACPI_error_status(Status);
		syslog(LOG_CRIT, "Error while executing %s", object_path);
		return NULL;
	}

	/* Ensure that the return object is large enough */
	if (ReturnValue.Length < sizeof (ACPI_OBJECT)) {
		syslog(LOG_DEBUG, "Return value from %s method too small, %.8X\n", object_path, ReturnValue.Length);
		return NULL;
	}

	/* Expect an package return value from execution of _PSD */
	Object = (ACPI_OBJECT*)ReturnValue.Pointer;	
	if (Object->Type != ACPI_TYPE_PACKAGE) {
		syslog(LOG_DEBUG, "Invalid return type from %s, %.2X\n", object_path, Object->Type);
		return NULL;
	}

	return Object;
}


void lookAtPSD(uint8 cpuid)
{
	const utf8_t			*_PSDPaths[8] = {"\\_PR_.CPU0._PSD", "\\_PR_.CPU1._PSD", "\\_PR_.CPU2._PSD", "\\_PR_.CPU3._PSD", "\\_PR_.CPU4._PSD", "\\_PR_.CPU5._PSD", "\\_PR_.CPU6._PSD", "\\_PR_.CPU7._PSD"};
	ACPI_OBJECT				*object;
	int i;

	/*Execute the _PCT method for the cpu - obtaining the object*/
    object = execute_acpi_object_with_return_type_package(_PSDPaths[cpuid]);
	if (object == NULL) {
		syslog(LOG_DEBUG, "package is NULL, execution failed for object %s ", _PSDPaths[cpuid]);
		return;
	}

	/*Validate the object*/
	if (  object->Type != ACPI_TYPE_PACKAGE) {
		syslog(LOG_DEBUG, "ERROR: Package returned from executing %s has incorrect format", _PSDPaths[cpuid]);
		return;
	}

	syslog(LOG_DEBUG, "The number of dependencies is: %d", object->Package.Count);

	for (i = 0; i < object->Package.Count; i++) {
		syslog(LOG_DEBUG, "Num Entries: %d (should be 5?)", object->Package.Elements[i].Package.Elements[0]);
		syslog(LOG_DEBUG, "Revision: %d (should be 0 ?)", object->Package.Elements[i].Package.Elements[1]);
		syslog(LOG_DEBUG, "Domain: %d", object->Package.Elements[i].Package.Elements[2]);
		syslog(LOG_DEBUG, "CoordType: %x", object->Package.Elements[i].Package.Elements[3]);
		syslog(LOG_DEBUG, "NumProcessors: %d", object->Package.Elements[i].Package.Elements[4]);
	}
}


void ExecutePciHostBridge_OSC(UINT32 dword1, UINT32 dword2, UINT32 dword3)
{
	ACPI_STATUS             Status;
    ACPI_OBJECT_LIST        ArgList;
    ACPI_OBJECT             Args[4];
    ACPI_BUFFER             ReturnValue;
    ACPI_OBJECT             *Object;
	UINT8					UUID[16] = {0x5B, 0x4D, 0xDB, 0x33, 0xF7, 0x1F, 0x1C, 0x40, 0x96, 0x57, 0x74, 0x41, 0xC0, 0x3D, 0xD7, 0x66};
	UINT32					Capabilities[3] = {dword1, dword2, dword3};														  
	UINT32					*ReturnedWords;


    /* Setup input arguments */
    ArgList.Count = 4;
    ArgList.Pointer = Args;

	Args[0].Type = ACPI_TYPE_BUFFER;
	Args[0].Buffer.Pointer = UUID;
	Args[0].Buffer.Length = sizeof(UUID);

	Args[1].Type = ACPI_TYPE_INTEGER;
	Args[1].Integer.Value = 1;
	
	Args[2].Type = ACPI_TYPE_INTEGER;
	Args[2].Integer.Value = 3; 

	Args[3].Type = ACPI_TYPE_BUFFER;
	Args[3].Buffer.Pointer = (uint8 *)Capabilities;
	Args[3].Buffer.Length = sizeof(Capabilities);

	/* Ask ACPICA to allocate space for the return object */
    ReturnValue.Length = ACPI_ALLOCATE_BUFFER;
	

	Status = AcpiEvaluateObject (NULL, "\\_SB_.PCI0._OSC", &ArgList, &ReturnValue);
    if (ACPI_FAILURE (Status)) {
		print_ACPI_error_status(Status);
		syslog(LOG_CRIT, "Error while executing \\_SB_.PCI0._OSC");
        return;
    }

	/* Ensure that the return object is large enough */
	if (ReturnValue.Length < sizeof (ACPI_OBJECT)) {
		syslog(LOG_DEBUG, "Return value from \\_SB_.PCI0._OSC method too small, %.8X\n", ReturnValue.Length);
		return;
	}

	/* Expect a buffer return value from execution of _OSI */
	Object = (ACPI_OBJECT *)ReturnValue.Pointer;
	if (Object->Type != ACPI_TYPE_BUFFER) {
		syslog(LOG_DEBUG, "Invalid return type from \\_SB_.PCI0._OSC, %.2X\n", Object->Type);
	}

	ReturnedWords = (uint32 *)(Object->Buffer.Pointer);
	syslog(LOG_DEBUG, "\\_PR_.CPU0._OSC returned Dword1: %x, Dword2 (Passed one way, disregard changes): %x, Dword3: %x", ReturnedWords[0], ReturnedWords[1], ReturnedWords[2]);

    AcpiOsFree (Object);
    return;  
}


/******************************************************************************
 *
 * FUNCTION:     ExecuteOSI
 *
 * PARAMETERS:  None
 *
 * RETURN:      None
 *
 * DESCRIPTION: This function queries the platform to check for support for an
 *				operating system given by a query string. If the given OS lacks
 *				support, this function will halt the booting sequence.
 *****************************************************************************/
void ExecuteOSI (void)
{
    ACPI_STATUS             Status;
    ACPI_OBJECT_LIST        ArgList;
    ACPI_OBJECT             Arg[1];
    ACPI_BUFFER             ReturnValue;
    ACPI_OBJECT             *Object;

    /* Setup input argument */
    ArgList.Count = 1;
    ArgList.Pointer = Arg;

    Arg[0].Type = ACPI_TYPE_STRING;
    Arg[0].String.Pointer = "Windows 2012";
    Arg[0].String.Length = strlen (Arg[0].String.Pointer);

    /* Ask ACPICA to allocate space for the return object */
    ReturnValue.Length = ACPI_ALLOCATE_BUFFER;

    Status = AcpiEvaluateObject (NULL, "\\_OSI", &ArgList, &ReturnValue);
    if (ACPI_FAILURE (Status)) {
		syslog(LOG_CRIT, "Error while executing _OSI");
        return;
    }

    /* Ensure that the return object is large enough */
    if (ReturnValue.Length < sizeof (ACPI_OBJECT)) {
        AcpiOsPrintf ("Return value from _OSI method too small, %.8X\n",
            ReturnValue.Length);
        return;
    }

    /* Expect an integer return value from execution of _OSI */
    Object = ReturnValue.Pointer;
    if (Object->Type != ACPI_TYPE_INTEGER)
        AcpiOsPrintf ("Invalid return type from _OSI, %.2X\n", Object->Type);

	if ((UINT32)Object->Integer.Value != 0xFFFFFFFF)
		syslog(LOG_CRIT, "OS not supported by platform");

	if (VORTEX_ACPI_DEBUG)
		syslog(LOG_DEBUG, "_OSI returned 0x%8.8X", (UINT32) Object->Integer.Value);

    AcpiOsFree (Object);
    return;
}


//Functions for working with C-states

cstate_array_t *getPowerSupportedStates(uint32 cpuid)
{
	const utf8_t			*_CSTPaths[8] = {"\\_PR_.CPU0._CST", "\\_PR_.CPU1._CST", "\\_PR_.CPU2._CST", "\\_PR_.CPU3._CST", "\\_PR_.CPU4._CST", "\\_PR_.CPU5._CST", "\\_PR_.CPU6._CST", "\\_PR_.CPU7._CST"};
	ACPI_OBJECT				*cst;
	cstate_array_t		    *c_state_array;
	uint32					num_cstates, i, cstate_count, cstate_index, acpi_address_space;

	if (cpuid < 0 || cpuid > 7) {
		syslog(LOG_DEBUG, "ERROR: In getPowerSupportedStates - cpuid out of range");
		goto error;
	}

    cst = execute_acpi_object_with_return_type_package(_CSTPaths[cpuid]);
	if (!CST_is_valid(cst))
		goto error;

	/*Find the number of Supported c-states. Remember that C1 allways is supported*/
	num_cstates = get_num_cstates(cst);

	/*Allocate memory for the supported C-states*/
	c_state_array = (cstate_array_t *)pm_alloc(sizeof(cstate_array_t) + sizeof(cstate_t) * num_cstates, VX_FALSE);
	c_state_array->num_elems = num_cstates;

	/*Copy the cstate information*/
	cstate_count = cst->Package.Elements[0].Integer.Value;
	cstate_index = 0;
	for (i = 1; i <= cstate_count; i++) {

		if (cst->Package.Elements[i].Type == ACPI_TYPE_PACKAGE) {
			/* Do all the sanity checking we can here*/
			if (cst->Package.Elements[i].Package.Count != 4)
				continue;

			if (cst->Package.Elements[i].Package.Elements[0].Type != ACPI_TYPE_BUFFER)
				continue;

			/* Copy the values*/
			if ( cst->Package.Elements[i].Package.Elements[1].Integer.Value < 1)
				continue;
			else if ( i == 1 && cst->Package.Elements[i].Package.Elements[1].Integer.Value != 1) {

				c_state_array->elems[cstate_index].type = 1;
				c_state_array->elems[cstate_index].latency = 1;	//Default values read from our HP voxo-blades
				c_state_array->elems[cstate_index].power = 1000;//Default values read from our HP voxo-blades
				c_state_array->elems[cstate_index].entry_method =CSTATE_ENTRY_METHOD_C1_HALT;
				c_state_array->elems[cstate_index].address = 0;
				c_state_array->elems[cstate_index].is_valid = 1;
				cstate_index++;
			} else {
				c_state_array->elems[cstate_index].type = cst->Package.Elements[i].Package.Elements[1].Integer.Value;
				c_state_array->elems[cstate_index].latency = cst->Package.Elements[i].Package.Elements[2].Integer.Value;
				c_state_array->elems[cstate_index].power = cst->Package.Elements[i].Package.Elements[3].Integer.Value;
			
			
				acpi_address_space = cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[3];
				if (acpi_address_space == ACPI_ADDRESSS_SPACE_FFH) {
				
					if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[4] ==0x00) {
                        c_state_array->elems[cstate_index].entry_method = CSTATE_ENTRY_METHOD_C1_HALT;
						c_state_array->elems[cstate_index].address = 0;
					} else if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[4] == 0x01 && cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[5] == 1) {
						syslog(LOG_CRIT, "Have C1 i/O then Halt - Currently not supported");
                        c_state_array->elems[cstate_index].entry_method = CSTATE_ENTRY_METHOD_C1_IO_THEN_HALT;
						c_state_array->elems[cstate_index].address = (uint64_t)cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[7];
					} else if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[3] == 0x01 && cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[5] == 2) {
						syslog(LOG_CRIT, "Have Native C-state Instruction - Currently not supported");
					}
				
				} else if (acpi_address_space == ACPI_ADDRESS_SPACE_SYSTEM_IO) {
                    c_state_array->elems[cstate_index].entry_method = CSTATE_ENTRY_METHOD_CSTATE_SYSTEM_IO;
					c_state_array->elems[cstate_index].address = (uint64_t)cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[7];
				}

				c_state_array->elems[cstate_index].is_valid = 1;
				cstate_index++;
			}
		}
	}

	AcpiOsFree(cst);
	return c_state_array;


error:
	return NULL;
}


void lookAtCST(uint8 cpuid)
{
	const utf8_t			*_CSTPaths[8] = {"\\_PR_.CPU0._CST", "\\_PR_.CPU1._CST", "\\_PR_.CPU2._CST", "\\_PR_.CPU3._CST", "\\_PR_.CPU4._CST", "\\_PR_.CPU5._CST", "\\_PR_.CPU6._CST", "\\_PR_.CPU7._CST"};
	ACPI_OBJECT				*object, *cst;
	int i, cstate_count;

	/*Execute the _PCT method for the cpu - obtaining the object*/
    cst = execute_acpi_object_with_return_type_package(_CSTPaths[cpuid]);

	if (cst == NULL) {
		syslog(LOG_DEBUG, "package is NULL, execution failed for object %s ", _CSTPaths[cpuid]);
		return;
	}

	if (!CST_is_valid(cst))
		goto cst_object_error;

	cstate_count = cst->Package.Elements[0].Integer.Value;
	syslog(LOG_DEBUG, "The number of listed c-states is: %d", cstate_count);

	for (i = 1; i <= cstate_count; i++) {

		if (cst->Package.Elements[i].Type == ACPI_TYPE_PACKAGE) {
			syslog(LOG_DEBUG, "Item %d", i);
			syslog(LOG_DEBUG, "Register info:");
			syslog(LOG_DEBUG, "Address Space id: %x", cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[3]);
			syslog(LOG_DEBUG, "Register bit width: %x", cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[4]);
			syslog(LOG_DEBUG, "Register bit offset: %d", cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[5]);
			syslog(LOG_DEBUG, "Access Size: %d", cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[6]);
			syslog(LOG_DEBUG, "Address: %X", (uint64_t)cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[7]);

			syslog(LOG_DEBUG, " ");
			if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[3] == 0x7F) {
				syslog(LOG_DEBUG, "Have Functional Fixed Hardware...");
				if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[4] ==0x00) {
					syslog(LOG_DEBUG, "Have C1 Halt");
				} else if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[4] == 0x01 && cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[5] == 1) {
					syslog(LOG_DEBUG, "Have C1 i/O then Halt - Currently not supported");
				} else if (cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[3] == 0x01 && cst->Package.Elements[i].Package.Elements[0].Buffer.Pointer[5] == 2) {
					syslog(LOG_DEBUG, "Have Native C-state Instruction - Currently not supported");
				}
			}
			

			syslog(LOG_DEBUG, " ");
			syslog(LOG_DEBUG, "Type: %d (should be  > 0 ?)", cst->Package.Elements[i].Package.Elements[1].Integer.Value);
			syslog(LOG_DEBUG, "Latency: %d", cst->Package.Elements[i].Package.Elements[2].Integer.Value);
			syslog(LOG_DEBUG, "Power: %d", cst->Package.Elements[i].Package.Elements[3].Integer.Value);
			
		} else {
			syslog(LOG_DEBUG, "Seems we have a weird type (%d) of element %d", cst->Package.Elements[i].Type, i); 
			if (object->Package.Elements[i].Type == ACPI_TYPE_INTEGER)
				syslog(LOG_DEBUG, "It was of integer type and has value %d", cst->Package.Elements[i].Integer.Value);
		}
	}

cst_object_error:
	return;

}



static bool_t CST_is_valid(ACPI_OBJECT *cst)
{
	int cstate_count;

	if (cst == NULL) {
		syslog(LOG_DEBUG, "ERROR: In validate_CST - given CST pointer is NULL");
		goto cst_object_error;
	}

	/*Validate the object*/
	if (  cst->Type != ACPI_TYPE_PACKAGE) {
		syslog(LOG_DEBUG, "ERROR: Package returned when executing _CST has incorrect format");

		/* There has to be at least 2 elements */
        if ( cst->Package.Count < 2) {
			syslog(LOG_DEBUG, "ERROR: There are not enough elements present in the cst");
			goto cst_object_error;
		}

		/* According to the example given in the ACPI 5 spec, the first element of the 
		 * package should be an integer listing the number of C-states described by 
		 * the CST object. Further, this is how it is implemented in chromiumOS 
		 * (https://chromium.googlesource.com/chromiumos/third_party/kernel-next/+/v3.4-rc7/drivers/acpi/processor_idle.c) 
		 */
        if (cst->Package.Elements[0].Type != ACPI_TYPE_INTEGER) {
			syslog(LOG_DEBUG, "ERROR: The first element of the CST object was not of integer type");
			goto cst_object_error;
		}

		cstate_count = cst->Package.Elements[0].Integer.Value;
		if (cstate_count < 1 || cstate_count != cst->Package.Count -1) {
			syslog(LOG_DEBUG, "ERROR: C-state count value does not match the length of the cstate-package");
			goto cst_object_error;
		}
	}

	return (bool_t)TRUE;

cst_object_error:
	return (bool_t)FALSE;
}

static int32 get_num_cstates(ACPI_OBJECT *cst)
{
	int32 cstate_count, c1_is_listed, i, type, num_cstates;

	if (cst == NULL) {
		syslog(LOG_DEBUG, "ERROR: In validate_CST - given CST pointer is NULL");
		return -1;
	}

	c1_is_listed = 0;
	cstate_count = cst->Package.Elements[0].Integer.Value;
	num_cstates = 0;
	for (i = 1; i <= cstate_count; i++) {
		
		if (cst->Package.Elements[i].Type == ACPI_TYPE_PACKAGE) {
			if (cst->Package.Elements[i].Package.Count != 4)
				continue;

			type = cst->Package.Elements[i].Package.Elements[1].Integer.Value;
			if (type == 1)
				c1_is_listed = 1;

			num_cstates++;
		}
	}

	if (!c1_is_listed)
		num_cstates++;

	return num_cstates;
}