#include <vortex/common.h>
#include "../../../arch/x86_64/cpu.h"
#include <vortex/acpi/load_balancing_scheduler.h>
#include <vortex/cpumux.h>

lbs_t           *lbs;
static objref_t load_reporting_thread_ref;
static bool_t   unit_test_load_balancing_scheduler;
static uint32   CORE_UTILIZATIONS[8];
static uint64   CORE_CYCLES[8];
static uint64   CORE_TIMESTAMPS[8];

//#define DEBUG_LOAD_BALANCING_SCHEDULER

//#define PROFILE_LOAD_BALANCING_SCHEDULER
#ifdef PROFILE_LOAD_BALANCING_SCHEDULER

    #define PROFILE_LOAD_BALANCING_SCHEDULER_NUM_SAMPLES    10000
    uint64      unretire_samples[PROFILE_LOAD_BALANCING_SCHEDULER_NUM_SAMPLES];
    uint64      get_cpu_samples[PROFILE_LOAD_BALANCING_SCHEDULER_NUM_SAMPLES];
    uint64      retire_samples[PROFILE_LOAD_BALANCING_SCHEDULER_NUM_SAMPLES];
    uint64      set_inactive_samples[PROFILE_LOAD_BALANCING_SCHEDULER_NUM_SAMPLES];
    
    uint32      profiling_num_samples = 0;
    bool_t      is_first = VX_TRUE;
    filedesc_t  *scheduler_profiling_filedescs[4];
    utf8_t      *scheduler_profiling_fpaths[4] = {"root:/fs/scheduler_unretire_profiling.txt", 
                                                  "root:/fs/scheduler_get_cpu_profiling.txt", 
                                                  "root:/fs/scheduler_retire_profiling.txt", 
                                                  "root:/fs/scheduler_set_inactive_profiling.txt"
                                                 };

#endif

cpu_proxy_t *cpu_proxy_create(uint32 cpu_id)
{
    cpu_proxy_t *cpu_proxy;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    assert(cpu_id >= 0 && cpu_id < CPU_DRAFTNUM);
#endif

    cpu_proxy = (cpu_proxy_t *)pm_alloc(sizeof(cpu_proxy_t), VX_FALSE);
    if (cpu_proxy == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate cpu_proxy");
        return NULL;
    }

    cpu_proxy->cp_cpu_id = cpu_id;
    
    return cpu_proxy;
}


static vxerr_t lbs_alloc(lbs_t *lbs)
{
    vxerr_t     vxerr;
    uint32      i;
    cpu_proxy_t *tmp, *head;

    bzero(lbs, sizeof(lbs_t));

    //Create the lists so that they can be populated later according to any scheme we chose
    head = lbs->lbs_active_cpus;
    for (i = 0; i < 4; i++) {
        tmp = cpu_proxy_create(i);
        if (tmp == NULL) {
            vxerr = VXERR_NOMEM;
            goto error;
        }
        VxL_ENQUEUE(head, tmp);
    }
    lbs->lbs_active_cpus = head;

    
    head = lbs->lbs_inactive_cpus;
    for (i = 1; i < 8; i++) {
        tmp = cpu_proxy_create(i);
        if (tmp == NULL) {
            vxerr = VXERR_NOMEM;
            goto error;
        }
        VxL_ENQUEUE(head, tmp);
    }
    lbs->lbs_inactive_cpus = head;
    
    return VXERR_OK;

error:
    syslog(LOG_CRIT, "Unable to create cpu_proxy_object");
    return vxerr;
}

type_t lbs_type = {
    .t_name = "lbs",
    .t_size = sizeof(lbs_t),
    .t_alloc = (t_alloc_t) lbs_alloc,
};


/******************************************************************************
 *
 * FUNCTION:     get_percent_utilized
 *
 * PARAMETERS:  cpu_proxy   Pointer to cpu_proxy structure (corresponding to 
 *                          a cpu core).
 *
 * RETURN:      load        An integer giving the load of the core given by 
 *                          'cpu_proxy' in percent.
 *
 * DESCRIPTION: This function allows lbs to obtain the current load of the 
 *              different cpu cores (in percent).
 *
 *****************************************************************************/
uint32 get_percent_utilized(cpu_proxy_t *cpu_proxy)
{
    int64  cycle_diff, time_diff, cycles, time;

    uint32  load;
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER   
    assert(cpu_proxy != NULL);
    assert(cpu_proxy->cp_cpu_id >= 0 && cpu_proxy->cp_cpu_id < 8); 
#endif

    cycles = (int64)cpumux_rqroot[cpu_proxy->cp_cpu_id]->rq_rusage_cycles;
    cycle_diff = cycles - (int64)CORE_CYCLES[cpu_proxy->cp_cpu_id];
    
    CORE_CYCLES[cpu_proxy->cp_cpu_id] = (uint64)cycles;

    time = arch_timestamp_cycles();
    time_diff = time - (int64)CORE_TIMESTAMPS[cpu_proxy->cp_cpu_id];
    CORE_TIMESTAMPS[cpu_proxy->cp_cpu_id] = (uint64)time;

    load = (cycle_diff*100) / time_diff;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (unit_test_load_balancing_scheduler) 
        return CORE_UTILIZATIONS[cpu_proxy->cp_cpu_id];
    else
#endif
        return load;
}       


/******************************************************************************
 *
 * FUNCTION:     wakeup_necessary
 *
 * PARAMETERS:  lbs         Pointer to lbs structure
 *
 * RETURN:      VX_TRUE		A core should be awoken.
 *				VX_FALSE    No need to wake a core.
 *
 * DESCRIPTION: This function allows lbs to test whether additional core should 
 *              be awoken to serve the current load.
 *
 *****************************************************************************/
static bool_t wakeup_necessary(lbs_t *lbs)
{
    uint32      sum;
    bool_t      retval;
    objref_t    lbsref;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
#endif

    retval = VX_FALSE;

    if (lbs->lbs_active_cpus == NULL) {
        retval = VX_TRUE;
        goto exit;
    } else {

        sum = 0;
        VxL_FOREACH(lbs->lbs_active_cpus, {
            sum += get_percent_utilized(entry);
        });
    
        if (sum / VxL_QSIZE(lbs->lbs_active_cpus) > WAKE_UP_THRESHOLD) {
            retval = VX_TRUE;
            goto exit;
        }
    }

exit:

    return retval;
}


/******************************************************************************
 *
 * FUNCTION:     round_robin_get_cpu
 *
 * PARAMETERS:  lbs         Pointer to lbs structure
 *
 * RETURN:      cpu_id		Id of a core the task should be scheduled on.
 *
 * DESCRIPTION: This function allows for O(1) retrieval of a cpu_id for task 
 *              scheduling.
 *
 *****************************************************************************/
static uint32 round_robin_get_cpu(lbs_t *lbs)
{
    uint32      cpu_id;
    cpu_proxy_t *head, *tmp;
    objref_t    lbsref;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
    assert(lbs->lbs_active_cpus != NULL);
#endif

    head = lbs->lbs_active_cpus;
    tmp = (cpu_proxy_t *)VxL_DEQUEUE(head);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    assert(tmp != NULL);
#endif
    
    cpu_id = tmp->cp_cpu_id;
    VxL_ENQUEUE(head, tmp);
    lbs->lbs_active_cpus = head;

    return cpu_id;
}

static uint32 get_cpu_cache_preserving(lbs_t *lbs, uint32 current_affinity)
{
    cpu_proxy_t *head, *tmp;
    lock_t      lock;
    objref_t    lbsref;
    uint32      cpu_id, i, initial_size;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
    assert(lbs->lbs_active_cpus != NULL);
    assert(current_affinity >= 0 && current_affinity < CPU_DRAFTNUM);
#endif

    VxO_LOCKNULL(&lock);
    VxO_REFOBJ(&lbsref, lbs);
    if (lbsref.ob_object != lbs)
        syslog(LOG_CRIT, "Something is very wrong, obrefs not corresponding");

    VxO_REFLOCK(&lock, &lbsref);

    initial_size = VxL_QSIZE(lbs->lbs_active_cpus);
    for (i = 0; i < initial_size; i++) {

        tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_active_cpus);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
        assert(tmp != NULL);
#endif  
     
        if (tmp->cp_cpu_id == current_affinity ) {
            VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
            cpu_id = current_affinity;
            goto done;
        } else {
            VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
        }
    }
   
    tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_active_cpus);
    cpu_id = tmp->cp_cpu_id;
    VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);

done:
    VxO_XUNLOCK(&lock);

    return cpu_id;
}


/******************************************************************************
 *
 * FUNCTION:     set_cores_retiring
 *
 * PARAMETERS:  lbs         Pointer to lbs structure
 *
 * RETURN:      None
 *
 * DESCRIPTION: This function sets under utilized cores to the 'retiring' 
 *              state. This means that a timestamp indicating when the 
 *              core entered this state is stored in the cpu_proxy object, 
 *              and this object is then moved to the retiring queue.
 *
 *****************************************************************************/
static void set_cores_retiring(lbs_t *lbs)
{
    uint32      i, initial_size;
    cpu_proxy_t *tmp;
    objref_t    lbsref;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
#endif

    initial_size = VxL_QSIZE(lbs->lbs_active_cpus);
    for (i = 0; i < initial_size; i++) {

        tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_active_cpus);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
        assert(tmp != NULL);
#endif
        if (get_percent_utilized(tmp) < RETIREMENT_LOAD_THRESHOLD) {
            tmp->cpu_retiring_started = arch_timestamp_usec();
            VxL_ENQUEUE(lbs->lbs_retiring_cpus, tmp);
        } else {
            VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
        }
    }
}


/******************************************************************************
 *
 * FUNCTION:     unretire_or_wake_up_cores
 *
 * PARAMETERS:  lbs         Pointer to lbs structure
 *
 * RETURN:      None
 *
 * DESCRIPTION: This function is used to unretire or wake up inactive cores 
 *              according to the current load experienced by the system. 
 *
 *              Because it has no overhead hardware wise, we would rather
 *              unretire a retiring core than wake up one that is sleeping. 
 *              Also, to make sure that we schedule the task on this core we 
 *              insert it first rather than last in the list of active cpus. 
 *              We untail to take the one that most recently was set retiring. 
 *              This should maximize the number of cores becoming inactive.
 *
 *              No retiring cpus, must wake one up. Also, to make sure that 
 *              we schedule the task on this core we insert it first rather 
 *              than last in the list of active cpus.
 *
 *****************************************************************************/
static void unretire_or_wake_up_cores(lbs_t *lbs)
{
    cpu_proxy_t *tmp;
    objref_t    lbsref;

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
#endif

    while (wakeup_necessary(lbs) ) {

        if (lbs->lbs_retiring_cpus != NULL) {
            tmp = (cpu_proxy_t *)VxL_UNTAIL(lbs->lbs_retiring_cpus);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
            assert(tmp != NULL);
#endif
        } else if (lbs->lbs_inactive_cpus != NULL) {
            tmp = (cpu_proxy_t *)VxL_DEQUEUE(lbs->lbs_inactive_cpus);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
            assert(tmp != NULL);
#endif     
        } else {
            break;
        }

        VxL_PUSH(lbs->lbs_active_cpus, tmp);
    }
}


/******************************************************************************
 *
 * FUNCTION:    set_cores_inactive
 *
 * PARAMETERS:  lbs         Pointer to lbs structure
 *
 * RETURN:      None
 *
 * DESCRIPTION: This function is used to set cores that have been retiring 
 *              sufficiently long inactive.
 *
 *****************************************************************************/
static void set_cores_inactive(lbs_t *lbs)
{
    uint64      current_time, time_retired;
    uint32      initial_size, i;
    cpu_proxy_t *tmp;


#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
#endif

    current_time = arch_timestamp_usec();

    initial_size = VxL_QSIZE(lbs->lbs_retiring_cpus);
    for (i = 0; i < initial_size; i++) {

        tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_retiring_cpus);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
        assert(tmp != NULL);
#endif
        time_retired = current_time - tmp->cpu_retiring_started;
        if (time_retired > RETIREMENT_TIME_THRESHOLD) {
            VxL_ENQUEUE(lbs->lbs_inactive_cpus, tmp);
        } else {
            VxL_ENQUEUE(lbs->lbs_retiring_cpus, tmp);
        }
    }
}


uint64 lbs_schedule(lbs_t *lbs)
{
    uint64      core_id;
    lock_t      lock;
    objref_t    lbsref;

    VxO_LOCKNULL(&lock);
    VxO_REFOBJ(&lbsref, lbs);

    VxO_REFLOCK(&lock, &lbsref);

#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
    if (lbs == NULL)
        syslog(LOG_CRIT, "lbs pointer is NULL");
#endif

    //If no cores are bellow the load limit we must either 
    //unretire or wake one up so as to not incur exessive latency.
    unretire_or_wake_up_cores(lbs);

    //Next, we must get a core for running the task on
    core_id = (uint64)round_robin_get_cpu(lbs);

    //Finally, must set any underutilized cores as retiring
    //so power can be saved by eventually powering them down
    //TODO: Should we do something with the selected core so 
    //that it is not retired emidiately (If it was unretired)?
    set_cores_retiring(lbs);

    set_cores_inactive(lbs);

    VxO_XUNLOCK(&lock);
    return core_id;
}


// Test code below
static lbs_t *test_create_lbs(void)
{
    int32 core_id;
    lbs = (lbs_t * )VxO_NEW(lbs);
    core_id = 0;
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 4);
    VxL_FOREACH(lbs->lbs_active_cpus, {assert(entry->cp_cpu_id == core_id);
        core_id++;
    });

    assert(VxL_QSIZE(lbs->lbs_inactive_cpus) == 4);
    VxL_FOREACH(lbs->lbs_inactive_cpus, {assert(entry->cp_cpu_id == core_id);
        core_id++;
    });

    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 0);

    syslog(LOG_DEBUG, "Test of create_lbs passed [v]");

    return lbs;
}

static void test_wakeup_necessary(lbs_t *lbs)
{
    cpu_proxy_t *tmp;

    CORE_UTILIZATIONS[0] = 1;
    CORE_UTILIZATIONS[1] = 1;
    CORE_UTILIZATIONS[2] = 1;
    CORE_UTILIZATIONS[3] = 1;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;
    assert(!wakeup_necessary(lbs));

    CORE_UTILIZATIONS[0] = 78;
    CORE_UTILIZATIONS[1] = 1;
    CORE_UTILIZATIONS[2] = 1;
    CORE_UTILIZATIONS[3] = 1;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;
    assert(!wakeup_necessary(lbs));

    CORE_UTILIZATIONS[0] = 75;
    CORE_UTILIZATIONS[1] = 75;
    CORE_UTILIZATIONS[2] = 75;
    CORE_UTILIZATIONS[3] = 75;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;
    assert(!wakeup_necessary(lbs));

    CORE_UTILIZATIONS[0] = 76;
    CORE_UTILIZATIONS[1] = 76;
    CORE_UTILIZATIONS[2] = 76;
    CORE_UTILIZATIONS[3] = 76;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;
    assert(wakeup_necessary(lbs));

    //Remember that we round downwards in C, must thus have load corresponding to +1 at each core
    CORE_UTILIZATIONS[0] = 79;
    CORE_UTILIZATIONS[1] = 75;
    CORE_UTILIZATIONS[2] = 75;
    CORE_UTILIZATIONS[3] = 75;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;
    assert(wakeup_necessary(lbs));

    tmp = lbs->lbs_active_cpus;
    lbs->lbs_active_cpus = NULL;
    assert(wakeup_necessary(lbs));
    lbs->lbs_active_cpus = tmp;


    syslog(LOG_DEBUG, "Test of wakeup_necessary passed [v]");
}

static void test_round_robin_get_cpu(lbs_t *lbs)
{
    uint32 core_id;
    //Make sure that the list is what we imagine
    core_id = 0;
    VxL_FOREACH(lbs->lbs_active_cpus, {assert(entry->cp_cpu_id == core_id);
                                        core_id++;});

    //The first round should match
    for (core_id = 0; core_id < 4; core_id++) {
        assert(round_robin_get_cpu(lbs) == core_id);
    }

    //And the ordering should be the same the second round...
    for (core_id = 0; core_id < 4; core_id++) {
        assert(round_robin_get_cpu(lbs) == core_id);
    }
    
    syslog(LOG_DEBUG, "Test of round_robin_get_cpu passed [v]");
}

static void test_set_cores_retiring(lbs_t *lbs)
{
    uint32 core_id;
    //Make sure that the list is what we imagine
    core_id = 0;
    VxL_FOREACH(lbs->lbs_active_cpus, {assert(entry->cp_cpu_id == core_id);
                                        core_id++;});

    //Assert that core with id 0 is set retiring
    CORE_UTILIZATIONS[0] = 49;
    CORE_UTILIZATIONS[1] = 75;
    CORE_UTILIZATIONS[2] = 75;
    CORE_UTILIZATIONS[3] = 75;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;

    set_cores_retiring(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 3);
    core_id = 1;
    VxL_FOREACH(lbs->lbs_active_cpus, { assert(entry->cp_cpu_id == core_id);
                                        core_id++;
                                      });

    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 1);
    core_id = 0;
    VxL_FOREACH(lbs->lbs_retiring_cpus, {assert(entry->cp_cpu_id == core_id);
                                        core_id++;});

    CORE_UTILIZATIONS[0] = 49;
    CORE_UTILIZATIONS[1] = 49;
    CORE_UTILIZATIONS[2] = 49;
    CORE_UTILIZATIONS[3] = 49;
    CORE_UTILIZATIONS[4] = 0;
    CORE_UTILIZATIONS[5] = 0;
    CORE_UTILIZATIONS[6] = 0;
    CORE_UTILIZATIONS[7] = 0;

    set_cores_retiring(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 0);
    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 4);

    syslog(LOG_DEBUG, "Test of set_cores_retiring passed [v]");
}

static void test_set_cores_inactive(lbs_t *lbs)
{
        
    int32       initial_size;
    uint32      i;
    cpu_proxy_t *tmp;

    //Start by making all some cores active
    initial_size = VxL_QSIZE(lbs->lbs_retiring_cpus);
    for (i = 0; i < initial_size; i++) {

        tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_retiring_cpus);
#ifdef DEBUG_LOAD_BALANCING_SCHEDULER
        assert(tmp != NULL);
#endif
        VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
    }


    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 4);

    //Initialize the cores to having loads matching them being active
    CORE_UTILIZATIONS[0] = 100;
    CORE_UTILIZATIONS[1] = 100;
    CORE_UTILIZATIONS[2] = 100;
    CORE_UTILIZATIONS[3] = 100;

    //Set two cores retiring with some time in between them, then assert 
    //gradually that they are made inactive
    CORE_UTILIZATIONS[0] = 49;
    set_cores_retiring(lbs);
    set_cores_inactive(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 3);
    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 1);
    assert(VxL_QSIZE(lbs->lbs_inactive_cpus) == 4);
    
    arch_delay_usec(RETIREMENT_TIME_THRESHOLD/2);
    CORE_UTILIZATIONS[1] = 49;
    set_cores_retiring(lbs);
    set_cores_inactive(lbs);
    arch_delay_usec(RETIREMENT_TIME_THRESHOLD/2);

    set_cores_retiring(lbs);
    set_cores_inactive(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 2);
    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) <= 1);
    assert(VxL_QSIZE(lbs->lbs_inactive_cpus) >= 5);

    arch_delay_usec( (RETIREMENT_TIME_THRESHOLD/2) + 10);
    set_cores_retiring(lbs);
    set_cores_inactive(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 2);
    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 0);
    assert(VxL_QSIZE(lbs->lbs_inactive_cpus) >= 6);

    CORE_UTILIZATIONS[2] = 49;
    CORE_UTILIZATIONS[3] = 49;
    set_cores_retiring(lbs);
    set_cores_inactive(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 0);
    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 2);
    assert(VxL_QSIZE(lbs->lbs_inactive_cpus) == 6);
    arch_delay_usec(RETIREMENT_TIME_THRESHOLD + 10);
    set_cores_retiring(lbs);
    set_cores_inactive(lbs);
    assert(VxL_QSIZE(lbs->lbs_active_cpus) == 0);
    assert(VxL_QSIZE(lbs->lbs_retiring_cpus) == 0);
    assert(VxL_QSIZE(lbs->lbs_inactive_cpus) == 8);

    syslog(LOG_DEBUG, "Test of set_cores innactive passed [v]"); 
}

static void dummy_thread(void)
{
    return;
}


static void load_reporting_thread(void)
{
    uint32      i, initial_size;
    cpu_proxy_t *tmp;
    uint32      load;
    lock_t      lock;
    objref_t    lbsref;

    uint32      vals[8];
    uint32      ids[8];

    unit_test_load_balancing_scheduler = VX_FALSE;

    VxO_LOCKNULL(&lock);
    VxO_REFOBJ(&lbsref, lbs);
    if (lbsref.ob_object != lbs)
        syslog(LOG_CRIT, "Something is very wrong, obrefs not corresponding");

    syslog(LOG_DEBUG, "Load reported thread started");
    while (1) {
        //Sleep 5 seconds
        thread_suspend_timeout(&load_reporting_thread_ref, 5000000);
 
        VxO_REFLOCK(&lock, &lbsref);
        initial_size = VxL_QSIZE(lbs->lbs_active_cpus);  
        for (i = 0; i < initial_size; i++) {

            tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_active_cpus);
            load = get_percent_utilized(tmp);
            vals[i] = load;
            ids[i] = tmp->cp_cpu_id;
            VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
        }
  
        syslog(LOG_DEBUG, "Active CPUs:");
        for (i = 0; i < initial_size; i++) {
            syslog(LOG_DEBUG, "Load for core %d is %d", ids[i], vals[i]);    
        }
        
        initial_size = VxL_QSIZE(lbs->lbs_retiring_cpus);  
        for (i = 0; i < initial_size; i++) {

            tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_retiring_cpus);
            load = get_percent_utilized(tmp);
            vals[i] = load;
            ids[i] = tmp->cp_cpu_id;
            VxL_ENQUEUE(lbs->lbs_retiring_cpus, tmp);
        }

        syslog(LOG_DEBUG, "Retiring CPUs:");
        for (i = 0; i < initial_size; i++) {
            syslog(LOG_DEBUG, "Load for core %d is %d", ids[i], vals[i]);    
        }
        
        initial_size = VxL_QSIZE(lbs->lbs_inactive_cpus);  
        for (i = 0; i < initial_size; i++) {

            tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_inactive_cpus);
            load = get_percent_utilized(tmp);
            vals[i] = load;
            ids[i] = tmp->cp_cpu_id;
            VxL_ENQUEUE(lbs->lbs_inactive_cpus, tmp);
        }
        VxO_XUNLOCK(&lock);

        
        syslog(LOG_DEBUG, "Inactive CPUs:");
        for (i = 0; i < initial_size; i++) {
            syslog(LOG_DEBUG, "Load for core %d is %d", ids[i], vals[i]);    
        }
        

    }
}

vxerr_t start_lbs_load_reporting_thread(void)
{
    vxerr_t     vxerr;

    //Create the thread
    vxerr = thread_create_fmt(&load_reporting_thread_ref, 10000, AFFINITY_NAFFINITY, KERNEL_CPUAREF, load_reporting_thread, "");
	if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to create load reporting thread: %s", strvxerr(vxerr));
        goto error;
	}

    vxerr = thread_resume(&load_reporting_thread_ref, 0);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to start load reporting thread: %s", strvxerr(vxerr));
        goto error;
	}
    
error:
    return vxerr;
}

static vxerr_t load_generator_thread(uint32 percent_load)
{
    uint32  work_time, work_time_per_burst, sleep_time_per_burst, num_bursts;
    uint64  current_time, burst_complete_time;

    num_bursts = 100;
    work_time = (1000000/100)*percent_load;
    work_time_per_burst = work_time/num_bursts;
    sleep_time_per_burst = (1000000 - work_time)/num_bursts;

    while (1) {

        burst_complete_time = arch_timestamp_usec() + work_time_per_burst;
        while (arch_timestamp_usec() < burst_complete_time) {

        }
        thread_suspend_timeout(NULL, sleep_time_per_burst); 

    }

    return VXERR_OK;
}

static vxerr_t load_administration_thread(uint32 interval, uint32 percent_per_thread)
{
    uint32      total_percent, i;
    objref_t    tcbref;
    vxerr_t     vxerr;
    uint64      total_wakeups;

    thread_suspend_timeout(NULL, 20000000);

    total_percent = 0;
    while (total_percent < 300) {
        
        arch_delay_usec(interval);

        vxerr = thread_create_fmt(&tcbref, 10000, AFFINITY_NAFFINITY, KERNEL_CPUAREF, load_generator_thread, "i", percent_per_thread);
        if (vxerr != VXERR_OK) 
            syslog(LOG_DEBUG, "Unable to create new load generator thread, total_percent = %d", total_percent);

        vxerr = thread_resume(&tcbref, THREAD_FLAG_NFLAG);
        if (vxerr != VXERR_OK) 
            syslog(LOG_DEBUG, "Unable to resume new load generator thread, total_percent = %d", total_percent);

        total_percent += percent_per_thread;
        syslog(LOG_DEBUG, "Total Percent %d", total_percent);
    }

    syslog(LOG_DEBUG, "Done creating increasing load =)");
    syslog(LOG_DEBUG, "Retirement load threshold: %d, Retirement time threshold: %d, Wake up threshold: %d", RETIREMENT_LOAD_THRESHOLD, RETIREMENT_TIME_THRESHOLD, WAKE_UP_THRESHOLD);
    total_wakeups = 0;
    for (i = 0; i < CPU_DRAFTNUM; i++) {
        syslog(LOG_DEBUG, "Num idle periods @ core %d: %D", i, cpulocal[i].cl_num_idle);
        total_wakeups += cpulocal[i].cl_num_idle;
    }
    syslog(LOG_DEBUG, "Total wakeups: %d", total_wakeups);

    return vxerr;
}


vxerr_t start_load_administration_thread(uint32 interval, uint32 percent_per_thread)
{
    vxerr_t     vxerr;
    objref_t    load_administration_ref;

    vxerr = thread_create_fmt(&load_administration_ref, 10000, AFFINITY_NAFFINITY, KERNEL_CPUAREF, load_administration_thread, "ii", interval, percent_per_thread);
    if (vxerr != VXERR_OK)
        syslog(LOG_CRIT, "Unable to create load administration thread");
    vxerr = thread_resume(&load_administration_ref, THREAD_FLAG_NFLAG);
    if (vxerr != VXERR_OK)
        syslog(LOG_CRIT, "Unable to start load administration thread");

    return vxerr;
}

vxerr_t test_lbs(void) 
{
    uint32 i, initial_size;
    cpu_proxy_t *tmp;

    unit_test_load_balancing_scheduler = VX_TRUE;

    //Assert that the structure looks the way we want
    lbs = test_create_lbs();

    //Assert that the logic determining whether or not we must wake cores work
    test_wakeup_necessary(lbs);

    //Assert that we retrieve active cores correctly in a round robin fashion
    test_round_robin_get_cpu(lbs);
    
    //Assert that cores are set retiring correctly
    test_set_cores_retiring(lbs);
    
    //Assert that cores are moved into the inactive queue correctly
    test_set_cores_inactive(lbs);
    
    //Assert that our scheme for pinning tasks to cores works
    //test_task_is_pinned();
    
    //Start by making all some cores active
    initial_size = VxL_QSIZE(lbs->lbs_retiring_cpus);
    for (i = 0; i < initial_size; i++) {
        tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_retiring_cpus);
        VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
    }
    initial_size = VxL_QSIZE(lbs->lbs_inactive_cpus);
    for (i = 0; i < initial_size; i++) {
        tmp = (cpu_proxy_t *)VxL_POP(lbs->lbs_inactive_cpus);
        VxL_ENQUEUE(lbs->lbs_active_cpus, tmp);
    }

    start_lbs_load_reporting_thread();
    
}





