#include <vortex/acpi/gphr.h>
#include <vortex/common.h>
#include <vortex/vpmem.h>


//#define DEBUG_GPHR

struct gphr {
        uint32  gphr_depth;
        uint32  gphr_start;
        uint32  gphr_end;
        uint32  gphr_elems[0];
};

static  uint32 power_of_ten_values[8] = { 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000};

/*** Internal Helper Functions ***/
static inline uint32 ten_to_the_power_of(uint32 power)
{
#ifdef DEBUG_GPHR
    assert(power >= 0 && power < 8);
#endif
    return power_of_ten_values[power];
}

/*** GPHR Functions ***/
gphr_t *gphr_create(uint32 depth) 
{
    gphr_t *gphr;

    gphr = (gphr_t *)pm_alloc(sizeof(gphr_t) + sizeof(uint32) * (depth + 1), VX_FALSE);
    if (gphr == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate GPHR");
        return NULL;
    }

    gphr->gphr_depth = depth + 1;
    gphr->gphr_start = 0;
    gphr->gphr_end = 0;

    return gphr;
}

vxerr_t gphr_destroy(gphr_t *gphr)
{
    vxerr_t vxerr;

#ifdef DEBUG_GPHR
    if (gphr == NULL) {
        syslog(LOG_CRIT, "when attempting to destroy gphr, gphr pointer is NULL");
        vxerr = VXERR_NEXIST;
        goto error;
    }
#endif

    pm_free((vaddr_t)gphr, gphr->gphr_depth*sizeof(uint32) + sizeof(gphr_t));
    vxerr = VXERR_OK;

error:
    return vxerr;
}


//For adding a value to the sliding windows 
//Global Phase History Register
vxerr_t gphr_add(gphr_t *gphr, uint32 value)
{
    vxerr_t vxerr;

#ifdef DEBUG_GPHR
    if (gphr == NULL) {
        syslog(LOG_CRIT, "when attempting to add to gphr, gphr pointer is NULL");
        vxerr = VXERR_NEXIST;
        goto error;
    }
#endif

    gphr->gphr_elems[gphr->gphr_end] = value;
    gphr->gphr_end = (gphr->gphr_end + 1) % gphr->gphr_depth;

    if (gphr->gphr_end == gphr->gphr_start)
        gphr->gphr_start = (gphr->gphr_start + 1) % gphr->gphr_depth;

    vxerr = VXERR_OK;

error:
    return vxerr;
}



//For getting the key from the Global Phase History Register
//This key is to be used to find an index from the key_to_index_table
uint32 gphr_get_key(gphr_t *gphr)
{
    uint32 key, i;
    int32 current_index;

#ifdef DEBUG_GPHR
    if (gphr == NULL)
        syslog(LOG_CRIT, "when attempting to get key from gphr, gphr pointer is NULL");
#endif

    key = 0;
    for (i = 1; i <= gphr->gphr_depth-1; i++) {
        current_index = (gphr->gphr_end + gphr->gphr_depth - i) % gphr->gphr_depth;
        key += gphr->gphr_elems[current_index] * ten_to_the_power_of(i-1);
    }

    return key;
}

//Gets the last observed phase from the Global 
//Phase History Register
uint32 gphr_get_last_observed_phase(gphr_t *gphr)
{
    int32 previous_index;

#ifdef DEBUG_GPHR
    if (gphr == NULL)
        syslog(LOG_CRIT, "when attempting to get key from gphr, gphr pointer is NULL");
#endif

    previous_index = (gphr->gphr_end + gphr->gphr_depth - 1) % gphr->gphr_depth;
    return gphr->gphr_elems[previous_index];
}

vxerr_t test_gphr(void)
{
    vxerr_t vxerr;
    gphr_t *gphr;
    uint32  i;

    //Test GPHR implementation
    gphr = gphr_create(8);

    for (i = 1; i < 10; i++) {
        vxerr = gphr_add(gphr, i);
        assert(gphr_get_last_observed_phase(gphr) == i);
        if (vxerr != VXERR_OK)
            syslog(LOG_CRIT, "Something went wrong when adding %d to gphr", i);

        if (i == 8)
            assert(gphr_get_key(gphr) == 12345678);
        else if (i == 9)
            assert(gphr_get_key(gphr) == 23456789);
    }

    for (i = 1; i < 10; i++) {
        vxerr = gphr_add(gphr, i);
        assert(gphr_get_last_observed_phase(gphr) == i);
        if (vxerr != VXERR_OK)
            syslog(LOG_CRIT, "Something went wrong when adding %d to gphr", i);

        if (i == 1)
            assert(gphr_get_key(gphr) == 34567891);
        else if (i == 2)
            assert(gphr_get_key(gphr) == 45678912);
        else if (i == 3)
            assert(gphr_get_key(gphr) == 56789123);
        else if (i == 4)
            assert(gphr_get_key(gphr) == 67891234);
        else if (i == 5)
            assert(gphr_get_key(gphr) == 78912345);
        else if (i == 6)
            assert(gphr_get_key(gphr) == 89123456);
        else if (i == 7)
            assert(gphr_get_key(gphr) == 91234567);
        else if (i == 8)
            assert(gphr_get_key(gphr) == 12345678);
        else if (i == 9)
            assert(gphr_get_key(gphr) == 23456789);
    }

    vxerr = gphr_destroy(gphr);
    if (vxerr != VXERR_OK)
        syslog(LOG_CRIT, "gphr_destroy failed");

    syslog(LOG_DEBUG, "GPHR tests passed, everything seems to work!");

    return vxerr;
}
