#include <vortex/acpi/gphtpredictor.h>
#include <vortex/acpi/gphr.h>
#include <vortex/acpi/pht.h>
#include <vortex/common.h>
#include <vortex/vpmem.h>
#include "../../../arch/x86_64/cpu.h"
#include <vortex/perfmon/perfmon.h>
#include <vortex/acpi/platform/vortex_acpi_pstates.h>


//#define USE_NAIVE_FORECAST 1

#define X_DIM   10000
#define Y_DIM   10000
objref_t        mem_to_uop_test_thread_tcbref;

struct gpht_predictor {
    gphr_t  *gpht_gphr;
    pht_t   *gpht_pht;
    uint32  gpht_prediction_key;
};

typedef union cpuid_arch_perfmon_leaf cpuid_arch_perfmon_leaf_t;  
union cpuid_arch_perfmon_leaf {
    uint32  capl_blob[4];
    struct {
        union {
            uint32 eax;
            struct {
                uint32 capl_arch_perfmon_version_id: 8;
                uint32 capl_num_gen_purpose_perfmon_counters_per_cpu: 8;
                uint32 capl_gen_purpose_perfmon_counter_bitwidth: 8;
                uint32 capl_ebx_vector_length: 8;
            };
        };
        union {
            uint32 ebx;
            struct {
                uint32 capl_na_core_cyle_event: 1;
                uint32 capl_na_instruction_retired_event: 1;
                uint32 capl_na_reference_cycles_event: 1;
                uint32 capl_na_llc_reference_event: 1;
                uint32 capl_na_llc_misses_event: 1;
                uint32 capl_na_branch_instruction_retired_event: 1;
                uint32 capl_na_branch_mispredict_retired_event: 1;
                uint32 capl_reserved_ebx: 25;
            };
        };
        union {
            uint32 ecx;
            struct {
                uint32 capl_ecx_reserved1: 30;
                uint32 capl_use_fixed_function_perf_countes: 1;
                uint32 capl_ecx_reserved2: 1;
            };
        };
        union {
            uint32 edx;
            struct {
                uint32 capl_num_fixed_function_perf_counters: 5;
                uint32 capl_fixed_function_perf_counter_bitwidth: 8;
                uint32 capl_reserved_edx: 19;
            };
        };
    };
};

/*** GPHT Predictor Functions ***/
gpht_predictor_t *gpht_predictor_create(uint32 gphr_depth, uint32 pht_numentries)
{
    gpht_predictor_t    *gpht_predictor;
    uint32              i;

    gpht_predictor = (gpht_predictor_t *)pm_alloc(sizeof(gpht_predictor), VX_FALSE);
    if (gpht_predictor == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate memory for gpht predictor");
        return NULL;
    }

    gpht_predictor->gpht_gphr = gphr_create(gphr_depth);
    if (gpht_predictor->gpht_gphr == NULL) {
        syslog(LOG_CRIT, "Unable to create gphr for gpht_predictor");
        pm_free((vaddr_t)gpht_predictor, sizeof(gpht_predictor_t));
        return NULL;
    }

    //Initialize by filling with pstate 1 (highest)
    for (i = 0; i < gphr_depth; i++)
        gphr_add(gpht_predictor->gpht_gphr, 1);

    gpht_predictor->gpht_prediction_key = gphr_get_key(gpht_predictor->gpht_gphr);

    gpht_predictor->gpht_pht = pht_create(pht_numentries);
    if (gpht_predictor->gpht_pht == NULL) {
        syslog(LOG_CRIT, "Unable to create pht for gpht_predictor");
        pm_free((vaddr_t)gpht_predictor, sizeof(gpht_predictor_t));
        return NULL;
    }

    return gpht_predictor;
}

vxerr_t gpht_predictor_destroy(gpht_predictor_t *gpht)
{
    vxerr_t vxerr;

#ifdef DEBUG_GPHT_PREDICTOR
    if (gpht == NULL) {
        syslog(LOG_CRIT, "In gpht_predictor_destroy, gpht pointer is NULL");
        vxerr = VXERR_NEXIST;
        goto error;
    }
#endif

    vxerr = gphr_destroy(gpht->gpht_gphr);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to destroy gphr");
    }

    vxerr = pht_destroy(gpht->gpht_pht);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to destroy pht");
    }

    pm_free((vaddr_t)gpht, sizeof(gpht_predictor_t));
    vxerr = VXERR_OK;

error:
    return vxerr;
}


//Update the predictor. This is done by:
//  1)  Storing the observed phase as the prediction that is to be given the next time 
//      the sequence of phases that resulted in the last prediction is encountered again.
//  2)  Updating the GPHR with the observed phase
vxerr_t gpht_predictor_update_predictor(gpht_predictor_t *gpht, uint32 observed_phase)
{
    vxerr_t vxerr;

#ifdef DEBUG_GPHT_PREDICTOR
    if (gpht == NULL)
        syslog(LOG_CRIT, "gpht pointer is NULL");
#endif

    //1)
    vxerr = pht_update_entry(gpht->gpht_pht, gpht->gpht_prediction_key, observed_phase); 
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to update prediction in PHT");
        goto error;
    }

    //2)
    vxerr = gphr_add(gpht->gpht_gphr, observed_phase);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to add prediction to GPHR");
        goto error;
    }

error:
    return vxerr;
}

//Get a predicted phase: To be used when performance level of CPU is to be predicted
uint32 gpht_predictor_get_prediction(gpht_predictor_t *gpht)
{
    uint32  predicted_phase, gphr_key;

#ifdef DEBUG_GPHT_PREDICTOR
    if (gpht == NULL)
        syslog(LOG_CRIT, "gpht_predictor_get_prediction - gpht pointer is NULL");
#endif

    //Retrieve the key
    gphr_key = gphr_get_key(gpht->gpht_gphr);

    //Check if we find a key in the dict of gphr series
    if (pht_key_is_present(gpht->gpht_pht, gphr_key)) {
        //Return the prediction
        predicted_phase = pht_get_prediction(gpht->gpht_pht, gphr_key);

    } else {
        //Get the last observed phase and predict this as next phase
        predicted_phase = gphr_get_last_observed_phase(gpht->gpht_gphr);
        //Remove least recently used entry
        pht_remove_least_recently_used_entry(gpht->gpht_pht);
        //Add new entry corresponding to the key, and set the current prediction to the last observed phase
        pht_add_new_entry(gpht->gpht_pht, gphr_key, predicted_phase);

    }

     //Set the key of the entry to be updated with actual observed key
    gpht->gpht_prediction_key = gphr_key;

    return predicted_phase;
}


vxerr_t enable_memory_boundedness_profiling(void)
{
    uint32  cpuid_leaf_architectural_performance_monitoring;
    cpuid_arch_perfmon_leaf_t arch_perfmon_leaf;
    vxerr_t vxerr;

    cpuid_leaf_architectural_performance_monitoring = 0x0A;
    asm_cpuid(cpuid_leaf_architectural_performance_monitoring, 0, (uint32 *)&arch_perfmon_leaf);

    //We need support for Architectural Monitoring 
    if (arch_perfmon_leaf.capl_arch_perfmon_version_id < 0) {
        syslog(LOG_DEBUG, "Architectural Performance monitoring is not supported");
        vxerr = VXERR_NOSUPP;
        goto error;
    }
    
    //Assert that we can utilize general pupose performance monitoring registers
    //This can be verified by checking that bit 30 of ecx is NOT set
    if (arch_perfmon_leaf.capl_use_fixed_function_perf_countes) {
        syslog(LOG_DEBUG, "Should use Fixed-Function performance counters - Not currently supported");
        vxerr = VXERR_NOSUPP;
        goto error;
    }

    //We need at least two general purpose performance monitoring registers
    if (arch_perfmon_leaf.capl_num_gen_purpose_perfmon_counters_per_cpu < 2) {
        syslog(LOG_DEBUG, "Not enough general purpose performance monitoring registers");
        vxerr = VXERR_NOSUPP;
        goto error;
    }
        
    // We need the ability to count last level hardware misses
    // This can be asserted in two steps:
    // 1) The ebx vector, enumerating Pre-defined architectural Performance events must be long enough
    //    to contain the LLC misses (index 4)
    // 2) The bit at index 4 must NOT be set, as this indicates lacking support for this feature
    if ( arch_perfmon_leaf.capl_ebx_vector_length < 5 || arch_perfmon_leaf.capl_na_llc_misses_event) {
        syslog(LOG_DEBUG, "LLC Miss performance monitoring events not available");
        vxerr = VXERR_NOSUPP;
        goto error;
    }

    //All is good now configure the Performance monitoring regs
    vxerr = arch_perfmon_config(PERFMON_COUNTER_0, PERFMON_EVTYPE_CACHE_L2_MISS, PERFMON_FLAG_COUNT_ALL);
    if (vxerr != VXERR_OK) {
        syslog(LOG_DEBUG, "Unable to configure Performance Monitoring Counter for LLC misses");
        goto error;

    }

    vxerr = arch_perfmon_config(PERFMON_COUNTER_1, PERFMON_EVTYPE_UOPS_RETIRED, PERFMON_FLAG_COUNT_ALL);
    if (vxerr != VXERR_OK) {
        syslog(LOG_DEBUG, "Unable to configure Performance Monitoring Counter for UOPS retired");
        goto error;
    } 

error:
    return vxerr;
}

uint32 predictor_classify_ratio(double mem_to_uops_ratio)
{
    uint32 phase;

    if ( mem_to_uops_ratio < 1.25)
        phase = 1;
    else if (mem_to_uops_ratio >= 1.25 && mem_to_uops_ratio < 1.75)
        phase = 2;
    else if (mem_to_uops_ratio >= 1.75 )
        phase = 3;

    return phase;
}

static double llc_miss_to_uops_ratio(void)
{
    uint64 llc_misses, uops_retired;
    lock_t lock;

    //Make certain that we are not interrupted in between reading these 
    //registers as this would mess up the ratio, making it useless as a memory 
    //boundedness metric
    VxO_INT_DISABLE(&lock);
    llc_misses = arch_perfmon_read(PERFMON_COUNTER_0);
    uops_retired = arch_perfmon_read(PERFMON_COUNTER_1);
    VxO_INT_ENABLE(&lock);

#ifdef DEBUG_PSTATE_BEHAVIOR
    if (CPU_ID == 3)
        fprintf(pstate_ratio_fd, "%D, %D\n", llc_misses, uops_retired);
#endif

    return (double)( (double)llc_misses/(double)uops_retired);
}

static void llc_get_values(uint64 *llc_misses, uint64 *uops_retired)
{
    lock_t lock;

    VxO_INT_DISABLE(&lock);
    *llc_misses = arch_perfmon_read(PERFMON_COUNTER_0);
    *uops_retired = arch_perfmon_read(PERFMON_COUNTER_1);
    VxO_INT_ENABLE(&lock);
}


void classify_phase(double mem_to_uops)
{
    if ( mem_to_uops < 0.005)
        syslog(LOG_DEBUG, "Phase 1 (highly CPU bound)");
    else if (mem_to_uops <= 0.005 && mem_to_uops < 0.010)
        syslog(LOG_DEBUG, "Phase 2 ");
    else if (mem_to_uops <= 0.010 && mem_to_uops < 0.015)
        syslog(LOG_DEBUG, "Phase 3 ");
    else if (mem_to_uops <= 0.015 && mem_to_uops < 0.020)
        syslog(LOG_DEBUG, "Phase 4 ");
    else if (mem_to_uops <= 0.020 && mem_to_uops < 0.030)
        syslog(LOG_DEBUG, "Phase 5 ");
    else if (mem_to_uops >= 0.030 )
        syslog(LOG_DEBUG, "Phase 6 (highly memory bound) ");
}

//Calculate primes, used as a cpu intensive e
static vxerr_t calculate_primes(uint32 upper_limit)
{
    double mem_to_uops, avg_mem_to_ups_ratio;
    uint32 i, num, primes;

    primes = 0;
    avg_mem_to_ups_ratio = 0;
    for (num = 1; num <= upper_limit; num++) { 

        for (i = 2; i <= num; i++) { 
            if(num % i == 0)
                break; 
        }
        if (i == num)
            primes++;

        if (primes % 1000 == 0) {
            mem_to_uops = llc_miss_to_uops_ratio();
            avg_mem_to_ups_ratio += mem_to_uops;
        }
    }

    return VXERR_OK;      
}


//Used as a very memory intensive test function
static vxerr_t matrix_multiply_reverse(void)
{
    double mem_to_uops, avg_mem_to_ups_ratio;
    uint32 i, j, num_iterations;
    uint64 **array;

    array = (uint64 **)pm_alloc(X_DIM*sizeof(uint64 *), VX_FALSE);
    if (array == NULL)
        syslog(LOG_CRIT, "Unable to allocate array...");
    
    for ( i = 0; i < X_DIM; i++) {
        array[i] = (uint64 *)pm_alloc(sizeof(uint64)*Y_DIM, VX_FALSE);
        if (array[i] == NULL)
            syslog(LOG_CRIT, "Unable to allocate array row number %d", i);
    }

    avg_mem_to_ups_ratio = 0;
    for (num_iterations = 0; num_iterations < 20; num_iterations++) {
        for (i = 0; i < X_DIM; i++) {
            for (j = 0; j < Y_DIM; j++) {
                array[j][i] += (i + j)*num_iterations;
            }
        }

        mem_to_uops = llc_miss_to_uops_ratio();
        avg_mem_to_ups_ratio += mem_to_uops;
        syslog(LOG_DEBUG, "Done with iteration %d of memory intensive calculations", num_iterations);
        classify_phase(mem_to_uops);
    }

    syslog(LOG_DEBUG, "Done running memory intensive code - Avg phase is:");
    classify_phase(avg_mem_to_ups_ratio);

    for ( i = 0; i < X_DIM; i++) {
        pm_free((vaddr_t)array[i], sizeof(uint64)*Y_DIM);
    }

    pm_free((vaddr_t)array, sizeof(uint64 *)*X_DIM);

    return VXERR_OK;
}

//Used as a memory intensive test function
static vxerr_t matrix_mutiply(void)
{
    double mem_to_uops, avg_mem_to_ups_ratio;
    uint32 i, j, num_iterations;
    uint64 **array;

    array = (uint64 **)pm_alloc(X_DIM*sizeof(uint64 *), VX_FALSE);
    if (array == NULL)
        syslog(LOG_CRIT, "Unable to allocate array...");
    
    for ( i = 0; i < X_DIM; i++) {
        array[i] = (uint64 *)pm_alloc(sizeof(uint64)*Y_DIM, VX_FALSE);
        if (array[i] == NULL)
            syslog(LOG_CRIT, "Unable to allocate array row number %d", i);
    }

    avg_mem_to_ups_ratio = 0;
    for (num_iterations = 0; num_iterations < 200; num_iterations++) {
        for (i = 0; i < X_DIM; i++) {
            for (j = 0; j < Y_DIM; j++) {
                array[i][j] += (i + j)*num_iterations;
            }
        }

        if (num_iterations % 10 == 0) {
            mem_to_uops = llc_miss_to_uops_ratio();
            avg_mem_to_ups_ratio += mem_to_uops;
            syslog(LOG_DEBUG, "Done with iteration %d of medium memory intensive calculations", num_iterations);
            classify_phase(mem_to_uops);
        }
    }

    syslog(LOG_DEBUG, "Done running medium memory intensive code - Avg phase is:");
    classify_phase(avg_mem_to_ups_ratio);

    for ( i = 0; i < X_DIM; i++) {
        pm_free((vaddr_t)array[i], sizeof(uint64)*Y_DIM);
    }

    pm_free((vaddr_t)array, sizeof(uint64 *)*X_DIM);

    return VXERR_OK;
}

static uint32 gpht_predictor_get_current_phase(void)
{
    return cpulocal[CPU_ID].cl_current_performance_state + 1;
}

static vxerr_t gpht_predictor_change_performance_state(uint32 desired_phase)
{
    vxerr_t vxerr;
    
    vxerr = acpi_change_performance_state(CPU_ID, 
                                          cpulocal[CPU_ID].cl_acpi_pstate_info->api_acpi_pstate_array , 
                                          desired_phase -1, 
                                          cpulocal[CPU_ID].cl_acpi_pstate_info->api_use_ffh, 
                                          cpulocal[CPU_ID].cl_acpi_pstate_info->api_control_register, 
                                          cpulocal[CPU_ID].cl_acpi_pstate_info->api_status_register);
 
    if (vxerr == VXERR_OK)
        cpulocal[CPU_ID].cl_current_performance_state = desired_phase -1;

    return vxerr;
}


static vxerr_t pstate_selector_thread(void)
{
    double  mem_to_uops_ratio;
    uint32  observed_phase, predicted_phase;
    vxerr_t vxerr;
    uint64  count;
    

    vxerr = enable_memory_boundedness_profiling();
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "enabling memory boundedness profiling failed");
    }

    cpulocal[CPU_ID].cl_pstate_selector_thread_active = VX_TRUE;
    count = 0;
    while (1) {
        
        predicted_phase = gpht_predictor_get_prediction(cpulocal[CPU_ID].cl_gpht_predictor);
        if (predicted_phase != gpht_predictor_get_current_phase()) {
            vxerr = gpht_predictor_change_performance_state(predicted_phase);
        }

        //Sleep 100 milliseconds
        thread_suspend_timeout(&cpulocal[CPU_ID].cl_pstate_selector_thread_ref, PSTATE_SELECTOR_SLEEPTIME);

        //Read out performance counter registers and update predictor accordingly
        mem_to_uops_ratio = llc_miss_to_uops_ratio();
        observed_phase = predictor_classify_ratio(mem_to_uops_ratio);

        vxerr = gpht_predictor_update_predictor(cpulocal[CPU_ID].cl_gpht_predictor, observed_phase); 
        if (vxerr != VXERR_OK)
            syslog(LOG_DEBUG, "Error when updating predictor");
    }
}

static vxerr_t naive_forecast_pstate_selector_thread(void)
{
    double  mem_to_uops_ratio;
    uint32  observed_phase, predicted_phase;
    vxerr_t vxerr;
    vxerr = enable_memory_boundedness_profiling();
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "enabling memory boundedness profiling failed");
    }



    cpulocal[CPU_ID].cl_pstate_selector_thread_active = VX_TRUE;
    observed_phase = 1;
    while (1) {
        
        predicted_phase = observed_phase;
        if (predicted_phase != gpht_predictor_get_current_phase()) {
            vxerr = gpht_predictor_change_performance_state(predicted_phase);
        }

        //Sleep 100 milliseconds
        thread_suspend_timeout(&cpulocal[CPU_ID].cl_pstate_selector_thread_ref, PSTATE_SELECTOR_SLEEPTIME);

        //Read out performance counter registers and update predictor accordingly
        mem_to_uops_ratio = llc_miss_to_uops_ratio();
        observed_phase = predictor_classify_ratio(mem_to_uops_ratio);
    }
}

vxerr_t start_pstate_selector_thread(uint32 cpuid)
{
    vxerr_t     vxerr;
    
    //Create the thread
#ifdef USE_NAIVE_FORECAST
    vxerr = thread_create_fmt(&cpulocal[cpuid].cl_pstate_selector_thread_ref, 10000, cpuid, KERNEL_CPUAREF, naive_forecast_pstate_selector_thread, "", NULL);
    syslog(LOG_DEBUG, "Starting Naive Forecast Prediction thread on core %d", cpuid);
#else
    vxerr = thread_create_fmt(&cpulocal[cpuid].cl_pstate_selector_thread_ref, 10000, cpuid, KERNEL_CPUAREF, pstate_selector_thread, "", NULL);
    syslog(LOG_DEBUG, "GPHT Prediction thread on core %d", cpuid);
#endif
	if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to create pstate selector thread: %s", strvxerr(vxerr));
            goto error;
	}

    vxerr = thread_resume(&cpulocal[cpuid].cl_pstate_selector_thread_ref, 0);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to start pstate selector thread thread: %s", strvxerr(vxerr));
        goto error;
	}
    
error:
    return vxerr;
}


static vxerr_t mem_to_uop_test_thread(void)
{
    double mem_to_uops;
    vxerr_t vxerr;
    thread_suspend_timeout(NULL, 5000000);

    vxerr = calculate_primes(100000);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Execution of CPU intensive code failed");
        goto error;
    }

    vxerr = matrix_mutiply();
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Execution of medium memory intensive code failed");
        goto error;
    }

    vxerr = matrix_multiply_reverse();
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Execution of memory intensive code failed");
        goto error;
    }

    while (1) {
        thread_suspend_timeout(NULL, 5000000);
        mem_to_uops = llc_miss_to_uops_ratio();
        classify_phase(mem_to_uops);
    }

error:
    return vxerr;
}


static vxerr_t start_mem_to_uop_test_thread(void)
{
    vxerr_t     vxerr;
    
    //Create the thread
    vxerr = thread_create_fmt(&mem_to_uop_test_thread_tcbref, 10000, AFFINITY_NAFFINITY, KERNEL_CPUAREF, mem_to_uop_test_thread, "", NULL);
	if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to create mem to uop timeout thread: %s", strvxerr(vxerr));
            goto error;
	}
	    
	//And start it
    vxerr = thread_resume(&mem_to_uop_test_thread_tcbref, 0);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Unable to start  mem to uop timeout thread: %s", strvxerr(vxerr));
        goto error;
	}
    
error:
    return vxerr;
}

vxerr_t test_gpht_predictor(void)
{
    vxerr_t vxerr;
    gpht_predictor_t *gpht_predictor;
    uint32 i;

    gpht_predictor = gpht_predictor_create(8, 3);
    assert(gpht_predictor != NULL);

    
    //Init by filling GPHR with ones
    for (i = 0; i < 8; i++) {
        gphr_add(gpht_predictor->gpht_gphr, 1);
    }

    //Assert that we get the previous value, that is, one
    assert(gpht_predictor_get_prediction(gpht_predictor) == 1);
    //Update the key '11111111' to predict the value 2 next time
    gpht_predictor_update_predictor(gpht_predictor, 2);

    //Check that the key now returned contains the two
    assert(gphr_get_last_observed_phase(gpht_predictor->gpht_gphr) == 2);

    //Flush the GPHR so that we can test whether two is now returned for a key of '11111111'
    for (i = 0; i < 7; i++) {
        gphr_add(gpht_predictor->gpht_gphr, 1);
    }
    //First test the edge case
    assert(gpht_predictor_get_prediction(gpht_predictor) == 1);
    gphr_add(gpht_predictor->gpht_gphr, 1);
    //And test for two after a full flush
    assert(gpht_predictor_get_prediction(gpht_predictor) == 2);

    vxerr = gpht_predictor_destroy(gpht_predictor);

    assert(vxerr == VXERR_OK);
    syslog(LOG_DEBUG, "test gpht predictor passed, all is good!");
    return vxerr;
}

