#include <vortex/acpi/pht.h>
#include <vortex/common.h>
#include <vortex/vpmem.h>

//#define DEBUG_PHT

struct pht_entry {
        uint32      pe_key;
        bool_t      pe_is_valid;
        uint32      pe_prediction;
        pht_entry_t *next;
        pht_entry_t *prev;
};

struct pht {
        dict_t          *pht_key_to_index_table;
        uint32          pht_numentries;
        //Pointer to list!
        struct pht_entry pht_entries;
};

type_t phtentry_type = {
    .t_name = "pht_entry",
    .t_size = sizeof(pht_entry_t),
    .t_alloc = KOBJ_ZEROFILL,
};


/*** PHT Functions ***/

pht_t *pht_create(uint32 numentries)
{
    pht_t *pht;
    uint32 i;
    pht_entry_t *tmp, *head;

    pht = (pht_t *)pm_alloc(sizeof(pht_t), VX_FALSE);
    if (pht == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate pht");
        return NULL;
    }

    pht->pht_numentries = numentries;

    //Create the dictionary
    pht->pht_key_to_index_table = VxD_NEW(SYNC_GLOBAL, DICT_TYPE_IKEY | DICT_TYPE_WEAKREF);
    if (pht->pht_key_to_index_table == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate dictionary for key to index table");
        return NULL;
    }

    //Create the list and populate it with nodes
    pht->pht_entries.prev = &pht->pht_entries;
    pht->pht_entries.next = (pht_entry_t *)NULL;
    head = &pht->pht_entries;

    for (i = 0; i < pht->pht_numentries; i++) {
        tmp = (pht_entry_t * )VxO_NEW(phtentry);
        if (tmp == NULL)
            syslog(LOG_CRIT, "Unable to create list pht entry");
        VxL_ENQUEUE(head, tmp);
    }

    return pht;
}

vxerr_t pht_destroy(pht_t *pht)
{
    vxerr_t     vxerr;
    pht_entry_t *head, *tmp;

#ifdef DEBUG_PHT
    if (pht == NULL) {
        vxerr = VXERR_NEXIST;
        goto error;
    }
#endif

    //Free the structure itself
    pm_free((vaddr_t)pht, sizeof(pht_t));
    vxerr = VXERR_OK;

error:
    return vxerr;
}

bool_t pht_key_is_present(pht_t *pht, uint32 key)
{
    vxerr_t     vxerr;
    objref_t    pht_entry_ref;

#ifdef DEBUG_PHT
    if (pht == NULL) {
        syslog(LOG_CRIT, "PHT pointer is null in pht_key_present");
    }

    assert(key >= 10000000);
#endif

    vxerr = VxD_GETITEM(&pht_entry_ref, pht->pht_key_to_index_table, (uid_t)key, NULL);
    if (vxerr == VXERR_OK)
        return TRUE;
    else
        return FALSE;
}

vxerr_t pht_update_entry(pht_t *pht, uint32 key, uint32 observed_phase)
{
    vxerr_t     vxerr;
    objref_t    pht_entry_ref;
    pht_entry_t *pht_entry;
    
#ifdef DEBUG_PHT
    if (pht == NULL) {
        vxerr = VXERR_NEXIST;
        goto error;
    }
    assert(key >= 10000000);
#endif

    //1)Lookup on the key
    vxerr = VxD_GETITEM(&pht_entry_ref, pht->pht_key_to_index_table, (uid_t)key, NULL);
    if (vxerr != VXERR_OK) {
        syslog(LOG_CRIT, "Attempting to get item with key that does not exist");
        goto error;
    }

    //2)follow the pointer to the correct node and 
    //  update the prediction with the observed phase
    pht_entry = (pht_entry_t *) VxO_REFGETOBJ(&pht_entry_ref);
    pht_entry->pe_prediction = observed_phase;
    vxerr = VXERR_OK;

error:
    return vxerr;
}

uint32 pht_get_prediction(pht_t *pht, uint32 key)
{
    vxerr_t     vxerr;
    objref_t    pht_entry_ref;
    pht_entry_t *pht_entry, *head;
    uint32      prediction;

#ifdef DEBUG_PHT
    if (pht == NULL)
        syslog(LOG_CRIT, "In pht_get_prediction, pht pointer is NULL");

    assert(key >= 10000000);
#endif

    //1)Lookup on the key
    vxerr = VxD_GETITEM(&pht_entry_ref, pht->pht_key_to_index_table, (uid_t)key, NULL);
    if (vxerr != VXERR_OK)
        syslog(LOG_CRIT, "Attempting to get item with key that does not exist");
    
    //2)Remove the node from the list, and put it at the back of the list
    //  updating its status to be the most recently used
    pht_entry = (pht_entry_t *) VxO_REFGETOBJ(&pht_entry_ref);
    head = &pht->pht_entries;
    VxL_QREMOVE(head, pht_entry);
    VxL_ENQUEUE(head, pht_entry);

    //3) Return the prediction
    return pht_entry->pe_prediction;;
}

vxerr_t pht_remove_least_recently_used_entry(pht_t *pht)
{
    pht_entry_t *lru, *head;
    vxerr_t     vxerr;
    objref_t    pht_entry_ref;

#ifdef DEBUG_PHT
    if (pht == NULL)
        syslog(LOG_CRIT, "PHT pointer is NULL");
#endif

    head = &pht->pht_entries;
    lru = head->next;
    VxL_QREMOVE(head, lru);

    if (lru->pe_is_valid) {
        //Remove the entry from the hashmap
        vxerr = VxD_GETITEM(&pht_entry_ref, pht->pht_key_to_index_table, (uid_t)lru->pe_key, NULL);
        if (vxerr != VXERR_OK)
            syslog(LOG_CRIT, "Attempting to get item with key that does not exist");
        vxerr = VxD_DELITEM(&pht_entry_ref, pht->pht_key_to_index_table, (uid_t)lru->pe_key, NULL);
        if (vxerr != VXERR_OK)
            syslog(LOG_CRIT, "Attempting to delete item that was just obtained failed");
    }
    else
        vxerr = VXERR_OK;

    //Free the memory of the entry
    VxO_DECREF(lru);
    
    return vxerr;
}

vxerr_t pht_add_new_entry(pht_t *pht, uint32 key, uint32 prediction)
{
    pht_entry_t *new_entry, *head;
    vxerr_t     vxerr;
    objref_t    pht_entry_ref;

#ifdef DEBUG_PHT
    //Removed as optimization
    if (pht == NULL) {
        syslog(LOG_CRIT, "In add_new_entry, pht pointer is NULL");
        vxerr = VXERR_NEXIST;
    }
    assert(key >= 10000000);
#endif

    //1) Create a new entry
    new_entry = (pht_entry_t *)VxO_NEW(phtentry);
    if (new_entry == NULL) {
        syslog(LOG_CRIT, "No memory, unable to allocate new pht entry");
        vxerr = VXERR_NOMEM;
        goto error;
    }

    //2)Populate it with key and prediction, set it valid
    new_entry->pe_is_valid = TRUE;
    new_entry->pe_key = key;
    new_entry->pe_prediction = prediction;

    //3)Enqueue it
    head = &pht->pht_entries;
    VxL_ENQUEUE(head, new_entry);

    //4)Add entry to the hashmap
    VxO_REFOBJ(&pht_entry_ref, new_entry);
    vxerr = VxD_SETITEM(&pht_entry_ref, pht->pht_key_to_index_table, (uid_t)key, NULL);
	if (vxerr == VXERR_EXIST) {
        syslog(LOG_CRIT, "Unable to add entry to hashmap");
		goto error;
    }
   
error:

    if (vxerr != VXERR_OK)
        syslog(LOG_CRIT, "About to leave pht_add_new_entry with non OK vxerr");

    return vxerr;
}

vxerr_t test_pht(void)
{
    vxerr_t vxerr;
    pht_t *pht;
    uint32 i;

    pht = pht_create(3);
    if (pht == NULL)
        syslog(LOG_CRIT, "Unable to create pht");

    assert(VxL_QSIZE(&pht->pht_entries) == 4 );

    //Remove least recently used item and check that list shrinks
    pht_remove_least_recently_used_entry(pht);
    assert(VxL_QSIZE(&pht->pht_entries) == 3 );

    //Add new entry, check that list size increases and that it becomes 
    //present in the dictionary
    pht_add_new_entry(pht, 10000000, 3);
    assert(VxL_QSIZE(&pht->pht_entries) == 4 );
    assert(pht_key_is_present(pht, 10000000));


    //Remove least recently used three times and add new ones.
    //Then assert that the previous key no longer is present 
    //but the three new ones are
    pht_remove_least_recently_used_entry(pht);
    pht_add_new_entry(pht, 10000001, 1);
    pht_remove_least_recently_used_entry(pht);
    pht_add_new_entry(pht, 10000002, 2);
    pht_remove_least_recently_used_entry(pht);
    pht_add_new_entry(pht, 10000003, 3);

    assert(pht_key_is_present(pht, 10000000) == FALSE);
    assert(pht_key_is_present(pht, 10000001));
    assert(pht_key_is_present(pht, 10000002));
    assert(pht_key_is_present(pht, 10000003));

    //Now test that our get prediction code works. Do this by
    //getting the prediction for the entry we added first and
    //verifying its value. Continue by removing the lru, and 
    //asserting that the one missing is the entry we added second.
    //Finally, add the one we deleted and assert they are all there.
    assert(pht_get_prediction(pht, 10000001) == 1);
    pht_remove_least_recently_used_entry(pht);
    assert(pht_key_is_present(pht, 10000002) == FALSE);
    assert(pht_key_is_present(pht, 10000001));
    assert(pht_key_is_present(pht, 10000003));
    pht_add_new_entry(pht, 10000002, 2);
    assert(pht_key_is_present(pht, 10000001));
    assert(pht_key_is_present(pht, 10000002));
    assert(pht_key_is_present(pht, 10000003));


    //Test that our update of predictions work
    assert(pht_update_entry(pht, 10000001, 5) == VXERR_OK);
    assert(pht_update_entry(pht, 10000002, 6) == VXERR_OK);
    assert(pht_update_entry(pht, 10000003, 7) == VXERR_OK);
    assert(pht_get_prediction(pht, 10000001) == 5);
    assert(pht_get_prediction(pht, 10000002) == 6);
    assert(pht_get_prediction(pht, 10000003) == 7);

    //Destroy the pht
    //assert(pht_destroy(pht) == VXERR_OK);
    syslog(LOG_DEBUG, "PHT tests passed, all is good!");
}

