#include <vortex/acpi/runlengthencoding.h>
#include <vortex/common.h>
#include <vortex/vpmem.h>

#define DEBUG_RLE


rle_t *create_rle(uint32 trace_length, uint32 origin_core)
{
	rle_t *rle;

#ifdef DEBUG_RLE
	assert(trace_length > 0);
	assert(trace_length % 2 == 0);
#endif

	rle = (rle_t *)pm_alloc(sizeof(rle_t) + sizeof(int64) * trace_length, VX_FALSE);
	if (rle == NULL) {
		syslog(LOG_CRIT, "Unable to allocate memory for runlength encoding");
		return NULL;
	}
	rle->num_elems = trace_length;
    rle->current_pos = 0;
    rle->samples_in_rle = 0;
    rle->origin_core = origin_core;
	return rle;
}

vxerr_t reset_rle(rle_t *rle)
{
    vxerr_t vxerr;
 
#ifdef DEBUG_RLE
    if (rle == NULL) {
		syslog(LOG_CRIT, "rle pointer is NULL");
		vxerr = VXERR_ILARG;
		goto error;
	}
#endif

    rle->current_pos = 0;
    rle->samples_in_rle = 0;
    vxerr = VXERR_OK;

error:
    return vxerr;
}

vxerr_t rle_add_sequence(rle_t *rle, int64 sequence_value, int64 sequence_length)
{
     vxerr_t vxerr;

 #ifdef DEBUG_RLE   
    assert(sequence_length >= 0);
    assert(rle->current_pos < rle->num_elems-1);

    if (rle == NULL) {
		syslog(LOG_CRIT, "rle pointer is NULL");
		vxerr = VXERR_ILARG;
		goto error;
	}
#endif

    //When we have sequences shorter than one microsecond
    if (sequence_length == 0)
        sequence_length = 1;

    rle->elems[rle->current_pos] = sequence_length;
    rle->elems[rle->current_pos + 1] = sequence_value;
    rle->current_pos += 2;
    rle->samples_in_rle += sequence_length;
    vxerr = VXERR_OK;
    
error:
    return vxerr;
}

vxerr_t destroy_rle(rle_t *rle)
{
	vxerr_t vxerr;

#ifdef DEBUG_RLE
	if (rle == NULL) {
		syslog(LOG_CRIT, "rle pointer is NULL");
		vxerr = VXERR_ILARG;
		goto error;
	}
#endif

	/* Because of our hacky approach we must free from the start of the array object and the number 
	 * of bytes to be freed corresponds to the number of items, their type, and a uint32 which was 
	 * used for holding the number of elements 
	 */
	pm_free((vaddr_t)rle, rle->num_elems*sizeof(int64) + sizeof(*rle));
	vxerr = VXERR_OK;

error:
	return vxerr;
}

bool_t rle_is_full(rle_t *rle)
{

#ifdef DEBUG_RLE
    if (rle == NULL) {
		syslog(LOG_CRIT, "rle pointer is NULL");
	}
#endif

    if (rle->current_pos == rle->num_elems)
        return TRUE;
    else
        return FALSE;
}

rle_t *copy_rle(rle_t *existing_rle)
{

#ifdef DEBUG_RLE
    if (existing_rle == NULL) {
		syslog(LOG_CRIT, "rle pointer is NULL");
        goto error;
	}
#endif

    rle_t *rle_copy = create_rle(existing_rle->num_elems, existing_rle->origin_core);
    if (rle_copy == NULL) {
        syslog(LOG_CRIT, "Unable to allocate new rle copy");
        goto error;
    }

    arch_bcopy(existing_rle, rle_copy, sizeof(rle_t) + sizeof(int64)*existing_rle->num_elems);

    return rle_copy;

error:
    return NULL;
}


