#include <vortex/acpi/sharemaster.h>
#include <vortex/acpi/runlengthencoding.h>
#include <vortex/acpi/fastexp.h>
#include <vortex/klib/math.h>
#include <vortex/common.h>
#include <vortex/rqueue.h>

//#define DEBUG_SHAREMASTER


uint32 cstate_hints[4] = {0xF0, 0x00, 0x11, 0x21};
double cstate_power_deltas[4] = {0.0, 8.0, 8.0, 9.25};
double cstate_entry_times[4] = {0.0, 10., 15.0, 17.0};
double cstate_exit_times[4] = {0.0, 10.0, 15.0, 17.0};

type_t share_master_type = {
	.t_name = "share_master_object",
	.t_size = sizeof(share_master_t),
	.t_alloc = KOBJ_ZEROFILL,
};

#ifdef ENABLE_SELECT_BEST_EXPERT_POLICY

share_master_t *create_share_master(double learning_rate, double alpha, uint32 num_experts, double c_wasting, double c_latency)
{
	share_master_t *master;
	vxerr_t			vxerr;
	assert(num_experts > 0 && num_experts <= MAX_STATIC_POLICIES);

	master = (share_master_t *)VxO_NEW(share_master);
	if (master == NULL) {
        syslog(LOG_CRIT, "Unable to create share_master object");
		goto error;
	}

	master->expert_cstates = (expert_cstate_array_t *)pm_alloc(sizeof(expert_cstate_array_t) + sizeof(uint32) * num_experts, VX_FALSE);
	if (master->expert_cstates == NULL) {
		syslog(LOG_CRIT, "Unable to allocate expert timeout array");
		goto error;
	}
	master->expert_cstates->num_elems = num_experts; 

	master->expert_weights = (expert_weight_array_t *)pm_alloc(sizeof(expert_weight_array_t) + sizeof(double) * num_experts, VX_FALSE);
	if (master->expert_cstates == NULL) {
		syslog(LOG_CRIT, "Unable to allocate expert weight array");
		goto error;
	}
	master->expert_weights->num_elems = num_experts;

	vxerr = _rq_new(&master->queueref, NULL, RESOURCE_METRIC_NONE, RQ_OWNTYPE_ANON, NULL);
	if (vxerr != VXERR_OK) {
		syslog(LOG_CRIT, "_rq_new() returned %s", strvxerr(vxerr));
		goto error;
	}

	master->learning_rate = learning_rate;
	master->alpha = alpha;
	master->num_experts = num_experts;
	master->c_wasting = c_wasting;
	master->c_latency = c_latency;
	return master;

error:
	return NULL;
}

void destroy_share_master(share_master_t *master)
{

#ifdef DEBUG_SHAREMASTER
	if (master == NULL) {
		syslog(LOG_CRIT, "share master pointer is NULL");
		return;
	}
#endif
	/* Because of our hacky approach we must free from the start of the array object and the number 
	 * of bytes to be freed corresponds to the number of items, their type, and a uint32 which was 
	 * used for holding the number of elements 
	 */
	pm_free((vaddr_t)master->expert_cstates, master->expert_cstates->num_elems*sizeof(uint32) + sizeof(*master->expert_cstates));
	pm_free((vaddr_t)master->expert_weights, master->expert_weights->num_elems*sizeof(double) + sizeof(*master->expert_weights));

	_rq_close(&master->queueref);
	 VxO_DECREF(master);
}

vxerr_t init_expert_weights(share_master_t *master)
{
	vxerr_t vxerr;
	uint32	i;

#ifdef DEBUG_SHAREMASTER
	if (master == NULL)
	{
		syslog(LOG_CRIT, "share master pointer is NULL");
		vxerr = VXERR_ILARG;
		goto done;
	}
#endif

	for (i = 0; i < master->num_experts; i++) {
		master->expert_weights->elems[i] = (double)(1.0/master->num_experts);
	}
	vxerr = VXERR_OK;

done:
	return vxerr;
}

uint32 calculate_best_expert(share_master_t *master)
{
    double highest_weight, best_expert;
    uint32 i;

#ifdef DEBUG_SHAREMASTER
	if (master == NULL) {
		syslog(LOG_CRIT, "share master pointer is NULL");
		return 0;
	}
#endif

    highest_weight = -1;
    best_expert = 0;
    for (i = 0; i < master->num_experts; i++) {

        if (master->expert_weights->elems[i] > highest_weight) {
            highest_weight = master->expert_weights->elems[i];
            best_expert = i;
        }
    }

    return best_expert;
}



static double power_saved_in_section(uint32 section_length, uint32 entry_time, uint32 exit_time, double power_delta)
{
    double power_saved;

    power_saved = 0.0;
    if (section_length > entry_time + exit_time) {
        power_saved = (double)((section_length - entry_time)*power_delta);
    }
    
    return power_saved;
}


static double get_best_power_saved(share_master_t *master, uint32 section_length)
{
    uint32 i, cstate;
    uint8_t pad[32];
    double saved, best_saved;
    uint8_t pad2[32];

#ifdef DEBUG_SHAREMASTER
	if (master == NULL) {
		syslog(LOG_CRIT, "share master pointer is NULL");
		return 0;
	}
#endif

    best_saved = 0;
    for (i = 0; i < master->num_experts; i++) {

        cstate = master->expert_cstates->elems[i];
        saved = power_saved_in_section(section_length, cstate_entry_times[cstate], cstate_exit_times[cstate], cstate_power_deltas[cstate]);
        if (saved > best_saved)
            best_saved = saved;
    }
      
    return best_saved;
}


double calculate_loss(share_master_t *master, rle_t *rle, uint32 cstate, uint32 entry_time, uint32 exit_time, double *best_value, bool_t do_sample_best)
{
	uint32  i, mode;
    double  power_saved, unpredicted, best_power_saved, saved_power_diff;
    int64   section_length, section_val;


#ifdef DEBUG_SHAREMASTER
	assert(cstate == 1 || cstate == 3);
    assert(entry_time == 10 || entry_time == 17);
    assert(exit_time == 10 || exit_time == 17);

	if (rle == NULL) {
		syslog(LOG_CRIT, "rle is NULL");
		goto error;
	}

	if (master == NULL) {
		syslog(LOG_CRIT, "share master pointer is NULL");
		goto error;
	}
#endif

	mode = SHARE_MASTER_CPU_MODE_ON;
	power_saved = 0;
    best_power_saved = 0;
	unpredicted = 0;
    
    for (i = 0; i + 1 < rle->current_pos; i += 2) {
        
		section_length = rle->elems[i];
		section_val = rle->elems[i+1];

		if (mode == SHARE_MASTER_CPU_MODE_OFF && section_val >= SHARE_MASTER_CPU_MODE_ON) {
            unpredicted += exit_time;
			mode = SHARE_MASTER_CPU_MODE_ON;
		
        } else if (mode >= SHARE_MASTER_CPU_MODE_ON && section_val == SHARE_MASTER_CPU_MODE_OFF) {
            power_saved += power_saved_in_section((uint32)section_length, entry_time, exit_time, cstate_power_deltas[cstate]);

            if (do_sample_best)
                best_power_saved += get_best_power_saved(master, (uint32)section_length);

            mode = SHARE_MASTER_CPU_MODE_OFF;    
		}
	}
    
    if (!do_sample_best)
        best_power_saved = *best_value;
    else 
        *best_value = best_power_saved;
  
    saved_power_diff = best_power_saved - power_saved;

    return (double)(master->c_wasting*saved_power_diff) + (double)(master->c_latency*unpredicted);

error:
	return -1.0;
}


vxerr_t update_expert_weights(share_master_t *master, rle_t *rle)
{

	double  expert_losses[MAX_STATIC_POLICIES];
	double  new_weights[MAX_STATIC_POLICIES];
	double  final_weights[MAX_STATIC_POLICIES];
	double  pool, weight_sum, best_value;
	int32   i, cstate;
      
	vxerr_t vxerr;

#ifdef DEBUG_SHAREMASTER
	if (master == NULL) {
		syslog(LOG_CRIT, "share master pointer is NULL");
		vxerr = VXERR_ILARG;
		goto error;
	}

	if (rle == NULL) {
		syslog(LOG_CRIT, "rle pointer is NULL");
		vxerr = VXERR_ILARG;
		goto error;
	}
#endif

	/* 1) Start by calculating the losses for each expert*/
	for (i = 0; i < master->num_experts; i++) {
        cstate = master->expert_cstates->elems[i];

        if (i == 0)
            expert_losses[i] = calculate_loss(master, rle, cstate, cstate_entry_times[cstate], cstate_exit_times[cstate], &best_value, VX_TRUE);
        else
            expert_losses[i] = calculate_loss(master, rle, cstate, cstate_entry_times[cstate], cstate_exit_times[cstate], &best_value, VX_FALSE);   
        
    }
    
	/* 2) Reduce the weights of poorly performing experts*/
	for (i = 0; i < master->num_experts; i++)
		new_weights[i] = master->expert_weights->elems[i]*fast_exp(-1.0*master->learning_rate*expert_losses[i]);

	/* 3) Share some of the remaining weights*/
	pool = 0.0;
	for (i = 0; i < master->num_experts; i++)
		pool += new_weights[i]*(1.0 - pow(1.0 - master->alpha, expert_losses[i]));

	for (i = 0; i < master->num_experts; i++)
		final_weights[i] = pow(1.0 - master->alpha, expert_losses[i])*new_weights[i] + (1.0/master->num_experts)*pool;

	/* 4) Rescale the weights and update the masters expert weights*/
	weight_sum = 0.0;
	for (i = 0; i < master->num_experts; i++)
		weight_sum += final_weights[i];

	for (i = 0; i < master->num_experts; i++)
		master->expert_weights->elems[i] = final_weights[i]/weight_sum;

	vxerr = VXERR_OK;

error:
	return vxerr;
}

#endif